from algoritmia.algorithms.mst import prim
from algoritmia.data.iberia import iberia2d, km2d
from algoritmia.viewers.graph2d_mst_viewer import Graph2dMstViewer

if __name__ == "__main__":
    # g_msf = kruskal(iberia2d, km2d)
    g_msf = prim(iberia2d, km2d)

    gv = Graph2dMstViewer(iberia2d, g_msf, canvas_width=800, canvas_height=800, margin=50)
    gv.run()



import tkinter
from enum import Enum
from typing import Optional

from easypaint import EasyPaint
from math import sin, cos, pi, sqrt

from algoritmia.datastructures.graphs import Digraph
from algoritmia.utils import infinity

# TECLAS
# - 'return', 'escape': Terminar la ejecución del programa
# - 'p': Capturar imagen como edit_distance.eps
# - 't': Cambiar el contenido de los nodos (3 posibilidades)
# - 'a': Cambiar las aristas (Grafo de dependencias y Backpointers)

type Num = int | float
type Vertex = tuple[Num, Num]
type Score = int
type Decision = str


class RecMode(Enum):
    Classic = 0
    Optimized = 1


# -------------------------------------------------------------------------------------


def run(s: str, t: str, sol: list[Decision]) -> str:
    n = 0
    res = list(s)
    p = 0
    for d in sol:
        if d == 'I':
            res.insert(p, t[n])
            n += 1
            p += 1
        elif d == 'D':
            del res[p]
        else:
            if d == 'S':
                res[p] = t[n]
            p += 1
            n += 1
    res.insert(p, '|')
    return ''.join(res)


type Trellis = dict[Vertex, tuple[str, Score, Decision]]


def edit_distance(mode: RecMode, s: str, t: str) -> tuple[Digraph, Trellis, list[Vertex]]:
    type SParams = tuple[int, int]  # (m, n)

    def S_clasic(m: int, n: int) -> Score:
        if m == 0 and n == 0:
            return 0
        if (m, n) not in mem:
            if n == 0:
                mem[m, n] = S_clasic(m - 1, n) + 1, (m - 1, n), 'D'
            elif m == 0:
                mem[m, n] = S_clasic(m, n - 1) + 1, (m, n - 1), 'I'
            else:
                c, d = (0, '-') if s[m - 1] == t[n - 1] else (1, 'S')
                mem[m, n] = min((S_clasic(m - 1, n) + 1, (m - 1, n), 'D'),
                                (S_clasic(m, n - 1) + 1, (m, n - 1), 'I'),
                                (S_clasic(m - 1, n - 1) + c, (m - 1, n - 1), d))
        return mem[m, n][0]

    def S_opt(m: int, n: int) -> Score:
        if m == 0 and n == 0:
            return 0
        if (m, n) not in mem:
            if n == 0:
                mem[m, n] = S_opt(m - 1, n) + 1, (m - 1, n), 'D'
            elif m == 0:
                mem[m, n] = S_opt(m, n - 1) + 1, (m, n - 1), 'I'
            else:
                if s[m - 1] == t[n - 1]:
                    mem[m, n] = S_opt(m - 1, n - 1), (m - 1, n - 1), '-'
                else:
                    mem[m, n] = min((S_opt(m - 1, n) + 1, (m - 1, n), 'D'),
                                    (S_opt(m, n - 1) + 1, (m, n - 1), 'I'),
                                    (S_opt(m - 1, n - 1) + 1, (m - 1, n - 1), 'S'))
        return mem[m, n][0]

    def path_from(m: int, n: int) -> tuple[list[Decision], list[Vertex]]:
        sol = []
        path = [(m, n)]
        while (m, n) != (0, 0):
            _, (m, n), dec = mem[m, n]
            sol.append(dec)
            path.append((m, n))
        sol.reverse()
        return sol, path

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}
    if mode == RecMode.Classic:
        S_clasic(len(s), len(t))
    else:
        S_opt(len(s), len(t))

    sol, path = path_from(len(s), len(t))

    e = []
    names2 = {}
    p = 0
    for n in range(len(t) + 1):
        if n == 0:
            names2[0, 0] = list(s)
            p = 0
        else:
            a = names2[0, n - 1][:]
            a.insert(p, t[n - 1])
            p += 1
            names2[0, n] = a
        for m in range(1, len(s) + 1):
            a = names2[m - 1, n][:]
            del a[p]
            # p -= 1
            names2[m, n] = a

    names = {}
    if mode == RecMode.Optimized:
        for (m, n) in names2:
            k = m, n
            names2[k].insert(n, '|')
            names[k] = ''.join(names2[k]), infinity, None

    for m in range(len(s) + 1):
        for n in range(len(t) + 1):
            if (m, n) == (0, 0):
                names[m, n] = '|' + s, 0, None
            else:
                if (m, n) in mem:
                    score, (mp, np), decision = mem[m, n]
                    e.append(((mp, np), (m, n)))
                    names[m, n] = run(s, t, path_from(m, n)[0]), score, decision
    return Digraph(E=e), names, path


# -------------------------------------------------------------------------------------

def dist(p1, p2):
    return sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]))


def find_intersec_point(cx, cy, radius, p1x, p1y, p2x, p2y):
    dx = p2x - p1x
    dy = p2y - p1y

    A = dx * dx + dy * dy
    B = 2 * (dx * (p1x - cx) + dy * (p1y - cy))
    C = (p1x - cx) * (p1x - cx) + (p1y - cy) * (p1y - cy) - radius * radius

    det = B * B - 4 * A * C
    if A <= 0.0000001 or det < 0:
        raise Exception("No real solutions")
    if det == 0:
        # One solution.
        t = -B / (2 * A)
        return (p1x + t * dx, p1y + t * dy), None
    else:
        # Two solutions.
        t1 = (-B + sqrt(det)) / (2 * A)
        t2 = (-B - sqrt(det)) / (2 * A)
        return (p1x + t1 * dx, p1y + t1 * dy), (p1x + t2 * dx, p1y + t2 * dy)


# -------------------------------------------------------------------------------------

class ViewMode(Enum):
    X_Y = 0
    ROW_COL = 1


class TextMode(Enum):
    Cursor = 0
    D = 1
    Score = 2


class ArrowMode(Enum):
    BackPointers = 0
    Dependecies = 1


class EditDistanceViewer(EasyPaint):
    def __init__(self, mode: RecMode, source: str, target: str,
                 canvas_width: int = 400, canvas_height: int = 300,
                 margin: int = 15, vertexmode: ViewMode = ViewMode.X_Y,
                 background: str = 'white', node_size: Optional[float] = None,
                 colors: Optional[dict[Vertex, str]] = None):

        g, names, path = edit_distance(my_mode, my_source, my_target)
        self.node_size = node_size
        self.margin = margin
        self.colors = colors if colors is not None else {}
        self.rec_mode = mode
        self.names = names
        self.path = path
        self.source = source
        self.target = target
        self.text_mode = TextMode.Cursor
        self.arrow_mode = ArrowMode.Dependecies
        if not isinstance(g, Digraph) or \
                any([not isinstance(p, tuple) or len(p) != 2 or
                     not isinstance(p[0], int) and not isinstance(p[0], float) or
                     not isinstance(p[1], int) and not isinstance(p[1], float) for p in g.V]):
            raise TypeError("The graph must be an UndirectedGraph. Vertices must be tuples of two integers or floats")

        if vertexmode == ViewMode.ROW_COL:
            if any([not isinstance(p[0], int) for p in g.V]):
                raise TypeError("In this mode, vertices must be tuples of two integers")
            g = Digraph(V=[(v[1], -v[0]) for v in g.V], E=[((u[1], -u[0]), (v[1], -v[0])) for (u, v) in g.E])
            path2 = [(v, -u) for (u, v) in self.path]
            self.path = path2
            names2 = dict(((v, -u), vv) for ((u, v), vv) in self.names.items())
            self.names = names2

        self.g = g

        self.max_y = max(p[1] for p in self.g.V)
        self.max_x = max(p[0] for p in self.g.V)
        self.min_y = min(p[1] for p in self.g.V)
        self.min_x = min(p[0] for p in self.g.V)

        super().__init__()
        self.easypaint_configure(title=f'Edit Distance Viewer  [{self.rec_mode}]',
                                 background=background,
                                 size=(canvas_width, canvas_height))

        w, h = self.size
        w -= self.margin
        h -= self.margin
        sizex = self.max_x - self.min_x + 1
        sizey = self.max_y - self.min_y + 1
        cell_size = self.cell_size = min(w / sizex, h / sizey)

        self.m = ((w - cell_size * sizex) / 2 - self.min_x * cell_size + self.margin / 2,
                  (h - cell_size * sizey) / 2 - self.min_y * cell_size + self.margin / 2)
        if self.node_size is None:
            self.node_size = 50  # (w * h / len(self.g.V)) ** 0.5 / 12  # cell_size / 8

    def draw_arrow(self, x1, y1, x2, y2, l2, color='black', width=1, tag='arrow'):
        phi = 30 * pi / 180
        l1 = ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5

        cell_size = self.cell_size
        m = self.m

        x3 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) + (y1 - y2) * sin(phi))
        x4 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) - (y1 - y2) * sin(phi))

        y3 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) - (x1 - x2) * sin(phi))
        y4 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) + (x1 - x2) * sin(phi))

        self.create_line((x1 + 0.5) * cell_size + m[0], (y1 + 0.5) * cell_size + m[1],
                         (x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)
        self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         (x3 + 0.5) * cell_size + m[0], (y3 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)
        self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                         (x4 + 0.5) * cell_size + m[0], (y4 + 0.5) * cell_size + m[1],
                         color, width=width, capstyle=tkinter.ROUND, tag=tag)

    def draw_vertex_text_aux(self, u: Vertex):
        if u not in self.names: return
        if self.text_mode == TextMode.Cursor:
            msg, fs = self.names[u][0], 12
        elif self.text_mode == TextMode.D:
            msg, fs = f'D({u[0]},{u[1]})', 12
        else:
            msg, fs = f'{self.names[u][1]}', 18
        self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                         (u[1] + 0.5) * self.cell_size + self.m[1],
                         msg, fs, tag='text')
        if self.text_mode != TextMode.Score and self.arrow_mode == ArrowMode.BackPointers:
            self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                             (u[1] + 0.35) * self.cell_size + self.m[1],
                             f'{self.names[u][1]}', 10, tag='text')

    def draw_vertex_text(self):
        if self.node_size > 0:
            self.erase('text')
            if self.arrow_mode == ArrowMode.BackPointers:
                for u in self.g.V:
                    self.draw_vertex_text_aux(u)
            else:
                for m in range(len(self.source) + 1):
                    for n in range(len(self.target) + 1):
                        u = m, n
                        self.draw_vertex_text_aux(u)

    def draw_vertex(self):
        if self.node_size > 0:
            self.erase('vertex')
            if self.arrow_mode == ArrowMode.BackPointers:
                for u in self.g.V:
                    width = 4 if u in self.path else 1
                    self.create_filled_circle((u[0] + 0.5) * self.cell_size + self.m[0],
                                              (u[1] + 0.5) * self.cell_size + self.m[1],
                                              self.node_size, fill='gray95', width=width, tag='vertex')
            else:
                width = 1
                for m in range(len(self.source) + 1):
                    for n in range(len(self.target) + 1):
                        u = m, n
                        self.create_filled_circle((u[0] + 0.5) * self.cell_size + self.m[0],
                                                  (u[1] + 0.5) * self.cell_size + self.m[1],
                                                  self.node_size, fill='gray95', width=width, tag='vertex')
                # x1 = (u[0] + 0.5) * cell_size + m[0]
                # y1 = (u[1] + 0.5) * cell_size + m[1]
                # w = self.node_size
                # h = self.node_size*0.25
                # self.create_filled_rectangle(x1-w, y1-h, x1+w, y1+h,
                #                              fill='palegreen')

    def draw_arrows(self):
        self.erase('arrow')
        for u, v in self.g.E:
            if self.arrow_mode == ArrowMode.BackPointers:
                width = 4 if v in self.path else 1
                u, v = v, u
                p1, p2 = find_intersec_point(v[0], v[1], (self.node_size + width) / self.cell_size, u[0], u[1], v[0],
                                             v[1])
                _, (x, y) = min((dist(u, p1), p1), (dist(u, p2), p2))
                color = 'blue' if self.names[u][2] == '-' else 'red'
                self.draw_arrow(u[0], u[1], x, y, 0.1, color, width=width, tag='arrow')
            else:
                width = 0
                for m in range(len(self.source) + 1):
                    for n in range(len(self.target) + 1):
                        v = m, n
                        for ix, iy in [(-1, -1), (-1, 0), (0, -1)]:
                            u = v[0] + ix, v[1] + iy
                            if u[0] < 0 or u[1] < 0: continue
                            p1, p2 = find_intersec_point(v[0], v[1], (self.node_size + width) / self.cell_size,
                                                         u[0], u[1], v[0], v[1])
                            _, (x, y) = min((dist(u, p1), p1), (dist(u, p2), p2))
                            color = 'blue' if (ix, iy) == (-1, -1) and self.source[m - 1] == self.target[
                                n - 1] else 'red'
                            self.draw_arrow(u[0], u[1], x, y, 0.1, color, width=width, tag='arrow')
                            if self.rec_mode == RecMode.Optimized and color == 'blue':
                                break  # a 'blue' arrow forbids 'red' brothers
            self.tag_lower('arrow')

    def draw_labels(self):
        for m in range(0, len(self.source)):
            self.create_text((m + 1) * self.cell_size + self.m[0],
                             0.1 * self.cell_size + self.m[1], self.source[m], 22)
        for n in range(0, len(self.target)):
            self.create_text(0.1 * self.cell_size + self.m[0],
                             (n + 1) * self.cell_size + self.m[1], self.target[n], 22)

    def on_key_press(self, keysym):
        if keysym.lower() in ['return', 'escape']:
            self.close()
            return
        if keysym.lower() == 'p':  # photo
            self.save_eps('edit_distance.eps')
            return

        if keysym.lower() == 't':
            if self.text_mode == TextMode.Cursor:
                self.text_mode = TextMode.D
            elif self.text_mode == TextMode.D:
                self.text_mode = TextMode.Score
            else:
                self.text_mode = TextMode.Cursor
        elif keysym.lower() == 'a':
            if self.arrow_mode == ArrowMode.BackPointers:
                self.arrow_mode = ArrowMode.Dependecies
            else:
                self.arrow_mode = ArrowMode.BackPointers
        self.draw_arrows()
        self.draw_vertex()
        self.draw_vertex_text()
        self.tag_raise('text')

    def main(self):
        self.draw_arrows()
        self.draw_vertex()
        self.draw_vertex_text()
        self.draw_labels()


# -------------------------------------------------------------------------------------


if __name__ == '__main__':
    my_mode = RecMode.Classic
    my_source, my_target = 'costa', 'casa'

    viewer = EditDistanceViewer(my_mode, my_source, my_target,
                                canvas_width=1000, canvas_height=800, vertexmode=ViewMode.X_Y)
    viewer.run()



"""
Created on 19/11/2018

@author: David Llorens dllorens@uji.es
"""
import sys

from easypaint import EasyPaint

from algoritmia.datastructures.kdtrees import Axis, KDTree, KDNode, KDLeaf


class KDTreeViewer(EasyPaint):
    def __init__(self, kdtree: KDTree):
        EasyPaint.__init__(self)
        self.kdtree = kdtree

    def draw_kdtree(self, kdtree: KDTree, min_x: float, min_y: float, max_x: float, max_y: float):
        if isinstance(kdtree, KDLeaf):
            if kdtree.point is not None:
                self.create_point(kdtree.point[0], kdtree.point[1])
        elif isinstance(kdtree, KDNode):
            if kdtree.axis == Axis.X:
                self.create_line(kdtree.split_value, min_y, kdtree.split_value, max_y, 'red')
                self.draw_kdtree(kdtree.child1, min_x, min_y, kdtree.split_value, max_y)
                self.draw_kdtree(kdtree.child2, kdtree.split_value, min_y, max_x, max_y)
            else:
                self.create_line(min_x, kdtree.split_value, max_x, kdtree.split_value, 'red')
                self.draw_kdtree(kdtree.child1, min_x, min_y, max_x, kdtree.split_value)
                self.draw_kdtree(kdtree.child2, min_x, kdtree.split_value, max_x, max_y)
        else:
            raise TypeError("Wrong type on 'kdtree' parameter")

    def get_points(self, kdtree: KDTree) -> list[tuple[float, float]]:
        if isinstance(kdtree, KDLeaf):
            return [kdtree.point]
        elif isinstance(kdtree, KDNode):
            return self.get_points(kdtree.child1) + self.get_points(kdtree.child2)
        else:
            raise TypeError("Wrong type on 'kdtree' parameter")

    def on_key_press(self, keysym):
        if keysym in ['Return', 'Escape']:
            self.close()

    def main(self):
        points = self.get_points(self.kdtree)
        x = list(map(lambda p: p[0], points))
        y = list(map(lambda p: p[1], points))
        min_x, max_x = min(x), max(x)
        min_y, max_y = min(y), max(y)

        sizex = 800
        sizey = sizex * (max_y - min_y) / (max_x - min_x)

        b = (max_x - min_x) * 0.05  # el borde será el 5% del tamaño del eje x
        self.easypaint_configure(title='KDTreeViewer',
                                 background='white',
                                 size=(sizex, sizey),
                                 coordinates=(min_x - b, min_y - b, max_x + b, max_y + b))

        self.create_rectangle(min_x, min_y, max_x, max_y, 'red')
        self.draw_kdtree(self.kdtree, min_x, min_y, max_x, max_y)


if __name__ == '__main__':
    # sys.argv.append("kdtree1.txt")
    if len(sys.argv) != 2:
        print("Use: python3 kdtreeviewer.py <file-kdtree.txt>")
        sys.exit()

    with open(sys.argv[1]) as content_file:
        content = content_file.read()
    try:
        loaded_kdtree = eval(content)
        KDTreeViewer(loaded_kdtree).run()
    except Exception:
        print("ERROR: el fichero no contiene un kdtree es el formato requerido.")



"""
2024-09-30: Versión 2.0 - Por cambios en LabyrinthViewer
2021-10-22: Versión 1.0 - Versión inicial
@author: David Llorens (dllorens@uji.es)

Hereda de la clase LabyrinthViewer para añadir animación de color.
Se anima el color de los vértices según apareren en la lista marked_cells (en ese el órden).
"""
from algoritmia.datastructures.graphs import UndirectedGraph
from algoritmia.viewers.labyrinth_viewer import LabyrinthViewer

type Vertex = tuple[int, int]
type Path = list[Vertex]


class LabyrinthViewerColor(LabyrinthViewer):
    def __init__(self,
                 lab: UndirectedGraph[Vertex],
                 canvas_width: int = 400, canvas_height: int = 400,
                 margin: int = 10, wall_width: int = 2, delay_ms: int = 0,
                 vertex_painted_per_iteration: int = 1):
        LabyrinthViewer.__init__(self, lab, canvas_width, canvas_height, margin, wall_width)
        self.state = 0
        self.vertex_painted_per_iteration = vertex_painted_per_iteration
        self.tt = None
        self.marked_cells2 = []
        self.delay_ms = delay_ms

    def on_key_press(self, keysym):
        if keysym != 'Return': return
        if self.state == 0:
            self.state = 1
            self.erase(self.tt)
            self.after(0, lambda: self.anim(0, len(self.marked_cells2)))
        elif self.state == 1:
            self.close()

    def anim(self, pos, max_pos):
        mw = self.mw
        mh = self.mh
        i = pos
        while i < min(max_pos, pos + self.vertex_painted_per_iteration):
            cell, color = self.marked_cells2[i]
            cx = mw / 2 + cell[1] * self.cell_size + self.cell_size / 2
            cy = mh / 2 + cell[0] * self.cell_size + self.cell_size / 2
            tt = self.create_filled_rectangle(cx - self.cell_size / 2, cy - self.cell_size / 2,
                                              cx + self.cell_size / 2, cy + self.cell_size / 2,
                                              'black', color, width=0)
            self.tag_lower(tt)
            i += 1
        self.update()
        if i < max_pos:
            self.after(self.delay_ms, lambda: self.anim(i, len(self.marked_cells2)))
            self.update()

    def add_marked_cell(self, cell, color='red'):
        self.marked_cells2.append((cell, color))

    def main(self):
        self.title = 'Labyrinth Viewer (colored cells)'
        super().main()
        w, h = self.center
        self.tt = self.create_text(w, h, "Press 'Return' to colorize", 14, "C", "red")


if __name__ == '__main__':
    import colorsys
    from algoritmia.datastructures.queues import Fifo


    def int2col(n: int, maxint: int = 255) -> str:
        (r, g, b) = colorsys.hsv_to_rgb(float(n) / maxint * (300 / 360.0), 1.0, 1.0)
        return "#{:02x}{:02x}{:02x}".format(int(255 * r), int(255 * g), int(255 * b))


    def matriz_distancias_anchura(grafo: UndirectedGraph[Vertex],
                                  v_inicial: Vertex) -> dict[Vertex, int]:
        dist: dict[Vertex, int] = {}
        queue = Fifo()
        seen: set[Vertex] = set()
        queue.push((v_inicial, v_inicial))
        seen.add(v_inicial)
        dist[v_inicial] = 0
        while len(queue) > 0:
            u, v = queue.pop()
            dist[v] = dist[u] + 1
            for suc in grafo.succs(v):
                if suc not in seen:
                    seen.add(suc)
                    queue.push((v, suc))
        return dist


    e = [((4, 7), (4, 6)), ((4, 7), (4, 8)), ((1, 3), (0, 3)), ((1, 3), (1, 4)), ((4, 8), (4, 9)), ((3, 0), (2, 0)),
         ((3, 0), (4, 0)), ((2, 8), (2, 7)), ((2, 8), (1, 8)), ((2, 1), (2, 0)), ((2, 1), (2, 2)), ((0, 0), (1, 0)),
         ((1, 6), (1, 5)), ((1, 6), (2, 6)), ((3, 7), (3, 8)), ((3, 7), (3, 6)), ((2, 5), (1, 5)), ((2, 5), (2, 4)),
         ((0, 3), (0, 2)), ((4, 0), (4, 1)), ((1, 2), (0, 2)), ((1, 2), (1, 1)), ((4, 9), (3, 9)), ((3, 3), (2, 3)),
         ((3, 3), (4, 3)), ((2, 9), (3, 9)), ((2, 9), (1, 9)), ((4, 4), (3, 4)), ((4, 4), (4, 3)), ((3, 6), (3, 5)),
         ((2, 2), (3, 2)), ((4, 1), (4, 2)), ((1, 1), (1, 0)), ((0, 1), (0, 2)), ((3, 2), (3, 1)), ((2, 6), (2, 7)),
         ((4, 5), (4, 6)), ((0, 4), (0, 5)), ((0, 4), (1, 4)), ((3, 9), (3, 8)), ((0, 5), (0, 6)), ((0, 7), (0, 6)),
         ((0, 7), (1, 7)), ((4, 2), (4, 3)), ((0, 8), (0, 9)), ((3, 5), (3, 4)), ((1, 8), (1, 7)), ((0, 9), (1, 9)),
         ((2, 3), (2, 4))]

    g0: UndirectedGraph[Vertex] = UndirectedGraph(E=e)
    num_rows, num_cols = max(v[0] for v in g0.V) + 1, max(v[1] for v in g0.V) + 1
    cell_size = 40
    margin0 = 10

    # Laberinto en forma de grafo no dirigido

    lv = LabyrinthViewerColor(g0,
                              canvas_width=num_cols * cell_size + margin0 * 2,
                              canvas_height=num_rows * cell_size + margin0 * 2,
                              margin=margin0,
                              delay_ms=300,
                              vertex_painted_per_iteration=1)

    start = (0, 0)  # (num_rows//2, num_cols//2)
    matriz_dist = matriz_distancias_anchura(g0, start)
    maxvalue = max(matriz_dist.values())
    for (v, k) in sorted([(v, k) for (k, v) in matriz_dist.items()]):
        lv.add_marked_cell(k, int2col(v, maxvalue))

    lv.run()



import unittest
from algoritmia.algorithms.topological_sort import topological_sort
from algoritmia.datastructures.graphs import Digraph

class TestTopsorter(unittest.TestCase):
    def test_topsort(self):
        G = Digraph(E=[(0,1), (1,2), (2,3), (3,4), (4,5)])
        self.assertEqual(tuple(topological_sort(G)), tuple(range(6)))
        self.assertEqual(tuple(topological_sort(G)), tuple(range(6)))
        G = Digraph(E=[('C', 'C++'), ('C', 'Objective C'), ('C', 'Java'), ('C', 'C#'),
                       ('C++', 'C#'), ('C++', 'Java'),
                       ('Java', 'C#')])
        ts = tuple(topological_sort(G))
        print(ts)
        for (u,v) in G.E:
            self.assertTrue(ts.index(u) < ts.index(v))

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


from collections.abc import Iterator
from dataclasses import dataclass
from typing import Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions, bt_vc_solutions, min_solution

# Tipos  --------------------------------------------------------------------------

type Decision = int     # Número de monedas del tipo actual
type Score = int        # Total de monedas utilizado

# Queremos que una solución sea la secuencia de decisiones (nº de monedas) en forma de tupla:
type Solution = tuple[Decision, ...]

# - 'bt_solutions' y 'bt_vc_solutions' devuelven un Iterator con las DecisionSequence que
#   llegan a una solución.
# - Pero un objeto DecisionSequence no es una tupla de decisiones: debemos utilizar el método
#   'decisions()' de la clase DecisionSequence para obtener la tupla.

# --------------------------------------------------------------------------------

def coin_change_solutions_naif(v: tuple[int, ...], Q: int) -> Iterator[Solution]:
    class CoinChangeDS(DecisionSequence[Decision, None]):
        def is_solution(self) -> bool:
            pending = calc_pending(self)  # O(n)
            return len(self) == len(v) and pending == 0

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(v):
                pending = calc_pending(self)  # O(n)
                for num_coins in range(pending // v[n] + 1):
                    yield self.add_decision(num_coins)

    def calc_pending(ds: CoinChangeDS) -> int:
        # Se calcula a partir del problema original y la secuencia de decisiones actual
        return Q - sum(d * v[i] for i, d in enumerate(ds.decisions()))

    initial_ds = CoinChangeDS()
    for solution_ds in bt_solutions(initial_ds):
        yield solution_ds.decisions()  # Extraemos las decisiones del objeto solution_ds y las devolvemos


# --------------------------------------------------------------------------------

def coin_change_solutions(v: tuple[int, ...], Q: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    new_pending = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(new_pending))

    initial_ds = CoinChangeDS(Extra(Q))
    for solution_ds in bt_solutions(initial_ds):
        yield solution_ds.decisions()


ScoredSolution = tuple[Score, Solution]


def coin_change_best_solution(v: tuple[int, ...], Q: int) -> ScoredSolution | None:
    def f(solution: Solution) -> int:
        return sum(solution)

    all_solutions = coin_change_solutions(v, Q)
    return min_solution(all_solutions, f)


# --------------------------------------------------------------------------------


def coin_change_vc_solutions(v: tuple[int, ...], Q: int) -> Iterator[Solution]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(DecisionSequence[Decision, Extra]):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    new_pending = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(new_pending))

        def state(self) -> tuple[int, int]:
            return len(self), self.extra.pending

    initial_ds = CoinChangeDS(Extra(Q))
    for solution_ds in bt_vc_solutions(initial_ds):
        yield solution_ds.decisions()


# Programa principal --------------------------------------------------------------------------------


if __name__ == "__main__":
    v0, Q0 = (1, 2, 5, 10), 11
    print(f"Coins: {v0}")
    print(f"Quantity: {Q0}\n")

    # Naif version
    print("Basic versión (all solutions):")
    sol0 = None
    for sol0 in coin_change_solutions_naif(v0, Q0):
        print(f"\tSolution: {sol0}")
    if sol0 is None:
        print("\tThere are no solutions")

    # Basic version
    print("Basic versión (all solutions):")
    sol0 = None
    for sol0 in coin_change_solutions(v0, Q0):
        print(f"\tSolution: {sol0}")
    if sol0 is None:
        print("\tThere are no solutions")

    print('Basic versión (best solution from all solutions):')
    print(f"\tBest solution: {coin_change_best_solution(v0, Q0)}")

    # Visited control version
    print("Visited control version:")
    sol0 = None
    for sol0 in coin_change_vc_solutions(v0, Q0):
        print(f"\tSolution: {sol0}")
    if sol0 is None:
        print("\tThere are no solutions")




from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction

if __name__ == "__main__":
    weighted_edges = {('Castelló', 'Sagunt'): 43, ('Sagunt', 'València'): 50}

    # Construimos la función de pesos a partir del diccionario anterior
    wf = WeightingFunction(weighted_edges, symmetrical=True)
    # Ejemplos de uso:
    print(wf('Castelló', 'Sagunt'))  # Output: 43
    print(wf('Sagunt', 'Castelló'))  # Output: 43 (al utilizar 'symmetrical=True')

    # Podemos aprovechar las claves del diccionario para crear el grafo
    road_graph = UndirectedGraph(E=weighted_edges.keys())

    # Podemos considerar un grafo de carreteras con distancias (pesos en las aristas) como una tupla
    weighted_road_graph = (road_graph, wf)



#
# Problema del cambio
#  - La solución óptima es la que utiliza menos monedas
#

from algoritmia.utils import infinity

# ------------------------------------------------------------------------------------------

type Quantity = int  # La cantidad a devolver

type Decision = int  # Un número de de monedas
type Solution = list[Decision]

type Score = int | float  # La puntuación es el número de monedas utilizadas o, si no hay solución, +infinito
type Result = tuple[Score, Solution] | None

type SParams = tuple[Quantity, int]

# ------------------------------------------------------------------------------------------

def solve(Q: Quantity, v: list[int]) -> Result:
    def S(q: Quantity, n: int) -> Score:
        # Casos base
        if n == 0 and q == 0: return 0
        if n == 0 and q > 0: return infinity
        # Recursividad memoizada
        if (q, n) not in mem:
            #mem[q, n] = min(((S(q - d * v[n - 1], n - 1) + d, (q - d * v[n - 1], n - 1), d)
            #                for d in range(q // v[n - 1] + 1)), default=infinity)
            mem[q, n] = infinity, (-1, -1), -1
            for d in range(q // v[n - 1] + 1):
               q_prev, n_prev = q - d * v[n - 1], n - 1
               current_score = S(q_prev, n_prev) + d
               mem[q, n] = min(mem[q, n], (current_score, (q_prev, n_prev), d))
        return mem[q, n][0]

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}
    score = S(Q, len(v))
    if score == infinity: return None
    # Si hay solución, la recuperamos de mem
    decisions = []
    q, n = Q, len(v)
    while n > 0:
        _, (q, n), d = mem[q, n]
        decisions.append(d)
    decisions.reverse()
    return score, decisions

# ------------------------------------------------------------------------------------------

if __name__ == '__main__':
    Q0, v0 = 11, [1, 2, 5, 10]
    print("Instance:")
    print(f"  Q: {Q0}")
    print(f"  v: {v0}\n")

    print("Result:", solve(Q0, v0))



from collections.abc import Iterator, Callable

from algoritmia.datastructures.graphs import IGraph, Edge, WeightingFunction, Weight
from algoritmia.datastructures.prioritymaps import MinHeapMap
from algoritmia.datastructures.queues import Fifo
from algoritmia.utils import argmin, infinity

type Traverse[T] = Callable[[IGraph[T], T], Iterator[Edge[T]]]


def traverse_bf[T](graph: IGraph[T],
                   v_initial: T) -> Iterator[Edge[T]]:
    queue: Fifo[Edge[T]] = Fifo()  # Cola de aristas
    seen: set[T] = set()  # Conjunto de vértices vistos
    queue.push((v_initial, v_initial))  # Añadimos la arista fantasma inicial
    seen.add(v_initial)
    while len(queue) > 0:
        u, v = queue.pop()
        yield u, v  # Generamos una arista
        for suc in graph.succs(v):
            if suc not in seen:
                queue.push((v, suc))
                seen.add(suc)


def traverse_df[T](graph: IGraph[T],
                   v_initial: T,
                   preorder: bool = True) -> Iterator[Edge[T]]:
    def traverse_from(u: T, v: T) -> Iterator[Edge[T]]:
        seen.add(v)
        if preorder:
            yield u, v  # Generamos una arista (recorrido en preorden)
        for suc_v in graph.succs(v):
            if suc_v not in seen:
                yield from traverse_from(v, suc_v)
        if not preorder:
            yield u, v  # Generamos una arista (recorrido en postorden)

    seen: set[T] = set()
    return traverse_from(v_initial, v_initial)  # Arista fantasma inicial


# Con diccionario: O(|V|^2)
def traverse_dijkstra_dict[T](g: IGraph[T],
                              d: WeightingFunction[T],
                              v_initial: T) -> Iterator[Edge[T]]:
    D: dict[T, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[T, T] = {v_initial: v_initial}
    fixed: set[T] = set()
    while len(bp) > 0:  # O(|V|) veces
        v = argmin(bp.keys(), lambda u: D[u])  # O(|V|)
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        del bp[v]
        for suc_v in g.succs(v):  # O(|V|)
            if suc_v not in fixed and D[v] + d(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + d(v, suc_v)
                bp[suc_v] = v


# Con diccionario de prioridad: O(|V| + |E| log |V|)
def traverse_dijkstra_heapmap[T](g: IGraph[T],
                                 d: WeightingFunction[T],
                                 v_initial: T) -> Iterator[Edge[T]]:
    D: MinHeapMap[T, Weight] = MinHeapMap((v, infinity) for v in g.V)  # O(|V|)
    D[v_initial] = 0
    bp: dict[T, T] = {v_initial: v_initial}
    fixed: set[T] = set()
    while len(D) > 0:
        v, dv = D.extract_opt_item()  # O(log |V|), O(|V|) veces
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        for suc_v in g.succs(v):
            if suc_v not in fixed and dv + d(v, suc_v) < D[suc_v]:
                D[suc_v] = dv + d(v, suc_v)  # O(log |V|), O(|E|) veces
                bp[suc_v] = v


# Con diccionario: O(|V|^2)
# En la práctica el coste es O(|V|) para grafos densos y O(sqrt(|V|) para grafos dispersos.
def traverse_dijkstra_metric_dict[T](g: IGraph[T],
                                     wf: WeightingFunction[T],
                                     eu_dist: Callable[[T, T], float],  # Function: Euclidean distance
                                     v_initial: T,
                                     v_final: T) -> Iterator[Edge[T]]:
    D: dict[T, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[T, T] = {v_initial: v_initial}
    fixed: set[T] = set()
    while len(bp) > 0:
        v = argmin(bp.keys(), lambda u: D[u] + eu_dist(u, v_final))
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        if v == v_final:
            break
        del bp[v]
        for suc_v in g.succs(v):
            if suc_v not in fixed and D[v] + wf(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + wf(v, suc_v)
                bp[suc_v] = v


if __name__ == '__main__':
    from algoritmia.datastructures.graphs import UndirectedGraph

    type Vertex = tuple[int, int]

    my_edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((0, 1), (0, 2)),
                ((2, 0), (1, 0)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=my_edges)
    initial_vertex = (0, 0)

    print('traverse_bf', list(traverse_bf(my_graph, initial_vertex)))
    print('traverse_df', list(traverse_df(my_graph, initial_vertex)))

    # ---------------------------------------------------------------------------------------

    from algoritmia.data.iberia import iberia, km, coords2d

    type Vertex = str  # Los vértices de iberia son nombres de ciudades

    traverse_from_Madrid = traverse_dijkstra_dict(iberia, km, 'Madrid')
    # Muestra solo los vértices:
    print('traverse_dijkstra_dict', [v for u, v in traverse_from_Madrid])


    def eu_dist0(city_a: Vertex, city_b: Vertex) -> float:
        pos2d_a, pos2d_b = coords2d[city_a], coords2d[city_b]
        dx, dy = pos2d_a[0] - pos2d_b[0], pos2d_a[1] - pos2d_b[1]
        return (dx * dx + dy * dy) ** 0.5


    Madrid_to_Bilbao = traverse_dijkstra_metric_dict(iberia, km, eu_dist0, 'Madrid', 'Bilbao')
    # Muestra solo los vértices:
    print('traverse_dijkstra_metric_dict', [v for u, v in Madrid_to_Bilbao])



import unittest
from dataclasses import dataclass
from typing import Iterator, Self

from algoritmia.schemes.bab_scheme import BabDecisionSequence, bab_max_solve


class TestBAB_SOLUTIONS(unittest.TestCase):

    def test_knapsack_bab_max_solve(self):
        Score = int
        State = tuple[int, int]

        @dataclass
        class Extra:
            weight: int
            value: int

        class KnapsackBabDS(BabDecisionSequence[int, Extra, Score]):
            def calculate_opt_bound(self) -> Score:
                return self.extra.value + sum(values[len(self):])

            def calculate_pes_bound(self) -> Score:
                return self.extra.value + 0  # No coge ninguno

            def is_solution(self) -> bool:
                return len(self) == len(values)

            def successors(self) -> Iterator[Self]:
                n = len(self)

                if n < len(values):
                    if self.extra.weight + weights[n] <= capacity:
                        new_weight = self.extra.weight + weights[n]
                        new_value = self.extra.value + values[n]
                        yield self.add_decision(1, Extra(new_weight, new_value))
                    yield self.add_decision(0, self.extra)

            def state(self) -> State:
                return len(self), self.extra.weight

        weights = [ 42,  55,  93, 89, 98, 77]
        values =  [168, 110, 186, 89, 98, 77]
        capacity = 136
        initial_ds = KnapsackBabDS(Extra(0, 0))
        score, solution_ds = bab_max_solve(initial_ds)
        solution = score, solution_ds.decisions()

        self.assertEqual((354, (1, 0, 1, 0, 0, 0)), solution)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#
# Problema del cambio generalizado
#  - Las monedas tiene peso
#  - La solución óptima es la de menor peso
#

from algoritmia.utils import infinity

# ------------------------------------------------------------------------------------------

type Decision = int       # Un número de de monedas
type Solution = list[Decision]

type Score = int | float  # La puntuación es el número de monedas utilizadas o, si no hay solución, infinity

# Declaramos un tipo Result que valga para todas las implementaciones,
# tanto las que devuelven la solución como las que no. Hay tres posibles resultsados:
#   - None: No hay solución
#   - (Score, None): Hay solución, pero solo se devuelve la puntuación de la solución
#   - (Score, Solution): Hay solución y se devuelve la solución y su puntuación
type Result = tuple[Score, Solution | None] | None

type SParams = tuple[int, int]  # Una tupla con los parámetros de S: (q, n)

# ------------------------------------------------------------------------------------------

def coin_change_no_mem(v: list[int], w: list[int], Q: int) -> Result:
    def S(q: int, n: int) -> Score:
        # Casos base
        if n == 0 and q == 0: return 0
        if n == 0 and q > 0: return infinity

        # Recursividad
        #return min((S(q - d * v[n - 1], n - 1) + d * w[n - 1]
        #           for d in range(q // v[n - 1] + 1)), default=infinity)
        best_score = infinity
        for d in range(q // v[n - 1] + 1):
            q_previo, n_previo = q - d * v[n - 1], n - 1
            current_score = S(q_previo, n_previo) + d * w[n - 1]
            best_score = min(best_score, current_score)
        return best_score

    score = S(Q, len(v))
    if score == infinity: return None
    return score, None


def coin_change_mem(v: list[int], w: list[int], Q: int) -> Result:
    def S(q: int, n: int) -> Score:
        # Casos base
        if q == 0 and n == 0: return 0
        if q > 0 and n == 0: return infinity

        # Recursividad con memoización
        if (q, n) not in mem:
            #mem[q, n] = min((S(q - d * v[n - 1], n - 1) + d * w[n - 1]
            #                 for d in range(q // v[n - 1] + 1)), default=infinity)
            mem[q, n] = infinity
            for d in range(q // v[n - 1] + 1):
                q_previo, n_previo = q - d * v[n - 1], n - 1
                current_score = S(q_previo, n_previo) + d * w[n - 1]
                mem[q, n] = min(mem[q, n], current_score)
        return mem[q, n]

    mem: dict[SParams, Score] = {}
    score = S(Q, len(v))
    if score == infinity: return None
    return score, None


def coin_change_rec(v: list[int], w: list[int], Q: int) -> Result:
    def S(q: int, n: int) -> Score:
        # Casos base
        if q == 0 and n == 0: return 0
        if q > 0 and n == 0: return infinity

        # Recursividad con memoización
        if (q, n) not in mem:
            # mem[q, n] = min(((S(q - d * v[n - 1], n - 1) + d * w[n - 1], (q - d * v[n - 1], n - 1), d)
            #                 for d in range(q // v[n - 1] + 1)), default=infinity)
            mem[q, n] = infinity, (-1, -1), -1
            for d in range(q // v[n - 1] + 1):
                q_previo, n_previo = q - d * v[n - 1], n - 1
                current_score = S(q_previo, n_previo) + d * w[n - 1]
                mem[q, n] = min(mem[q, n], (current_score, (q_previo, n_previo), d))
        return mem[q, n][0]

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}

    # La puntuación (peso) de la solución óptima
    score = S(Q, len(v))
    if score == infinity: return None

    # Recupera las decisiones
    decisions: list[Decision] = []
    q, n = Q, len(v)
    while (q, n) != (0, 0):
        _, (q, n), d = mem[q, n]
        decisions.append(d)
    decisions.reverse()

    return score, decisions


def coin_change_iter(v: list[int], w: list[int], Q: int) -> Result:
    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}

    # Rellena mem[., 0]
    mem[0, 0] = 0, (-1, -1), -1
    for q in range(1, Q + 1):
        mem[q, 0] = infinity, (-1, -1), -1

    for n in range(1, len(v) + 1):
        # Rellena mem[., n]
        for q in range(0, Q + 1):
            #mem[q, n] = min((mem[q - d * v[n - 1], n - 1][0] + d * w[n - 1]
            #                 for d in range(q // v[n - 1] + 1)), default=infinity)
            mem[q, n] = infinity, (-1, -1), -1
            for d in range(q // v[n - 1] + 1):
                q_previo, n_previo = q - d * v[n - 1], n - 1
                current_score = mem[q_previo, n_previo][0] + d * w[n - 1]
                mem[q, n] = min(mem[q, n], (current_score, (q_previo, n_previo), d))

    # La puntuación (peso) de la solución óptima
    score = mem[Q, len(v)][0]
    if score == infinity: return None

    # Recupera las decisiones
    decisions: list[Decision] = []
    q, n = Q, len(v)
    while (q, n) != (0, 0):
        _, (q, n), d = mem[q, n]
        decisions.append(d)
    decisions.reverse()

    return score, decisions


def coin_change_iter_score(v: list[int], w: list[int], Q: int) -> Result:
    previous: list[Score] = [infinity] * (Q + 1)  # n = -1 (reserva memoria)
    current: list[Score] = [0] + [infinity] * Q   # n = 0
    for n in range(1, len(v) + 1):
        current, previous = previous, current
        for q in range(0, Q + 1):
            #current[q] = min((previous[q - d * v[n - 1]] + d * w[n - 1]
            #                 for d in range(q // v[n - 1] + 1)), default=infinity)
            current[q] = infinity
            for d in range(q // v[n - 1] + 1):
                current_score = previous[q - d * v[n - 1]] + d * w[n - 1]
                current[q] = min(current[q], current_score)
    score = current[Q]
    if score == infinity: return None
    return score, None


if __name__ == '__main__':
    Q0, v0, w0 = 7, [1, 2, 4], [2, 3, 9]
    print(f"Instance:\n  Q = {Q0}\n  v = {v0}\n  w = {w0}\n")
    print('Result:')
    print("  coin_change_no_mem:", coin_change_no_mem(v0, w0, Q0))
    print("  coin_change_mem:", coin_change_mem(v0, w0, Q0))
    print("  coin_change_rec:", coin_change_rec(v0, w0, Q0))
    print("  coin_change_iter:", coin_change_iter(v0, w0, Q0))
    print("  coin_change_iter_score:", coin_change_iter_score(v0, w0, Q0))



# coding: latin1

import unittest

from algoritmia.datastructures.queues import Fifo, Lifo


class TestFifo(unittest.TestCase):
    def setUp(self):
        self.a = Fifo()
        self.b = Fifo([1, 2, 3])
        self.c = Fifo()
        self.d = Fifo([1, 2, 3])

    def test_ctor(self):
        self.assertEqual(len(self.a), 0)
        self.assertEqual(len(self.b), 3)
        self.assertEqual(len(self.c), 0)
        self.assertEqual(len(self.d), 3)

    def test_push(self):
        self.a.push(1)
        self.assertEqual(len(self.a), 1)
        self.a.push(2)
        self.assertEqual(len(self.a), 2)
        v = self.a.pop()
        self.assertEqual(v, 1)
        self.b.push(1)
        self.assertEqual(len(self.b), 4)
        self.c.push(1)
        self.assertEqual(len(self.c), 1)
        self.d.push(1)
        self.assertEqual(len(self.d), 4)

    def test_pop(self):
        self.assertRaises(IndexError, self.a.pop)
        i = 1
        while len(self.b) > 0:
            v = self.b.pop()
            self.assertEqual(v, i)
            i += 1
        self.assertRaises(IndexError, self.c.pop)
        i = 1
        while len(self.d) > 0:
            v = self.d.pop()
            self.assertEqual(v, i)
            i += 1

    def test_top(self):
        self.assertRaises(IndexError, self.a.top)
        self.assertEqual(self.b.top(), 1)


class TestLifo(unittest.TestCase):
    def setUp(self):
        self.a = Lifo()
        self.b = Lifo([1, 2, 3])

    def test_ctor(self):
        a = Lifo()
        b = Lifo([1, 2, 3])
        self.assertEqual(len(a), 0)
        self.assertEqual(len(b), 3)

    def test_push(self):
        self.a.push(1)
        self.assertEqual(len(self.a), 1)
        self.a.push(2)
        self.assertEqual(len(self.a), 2)
        v = self.a.pop()
        self.assertEqual(v, 2)

        self.assertEqual(len(self.b), 3)
        self.b.push(1)
        self.assertEqual(len(self.b), 4)

    def test_pop(self):
        self.assertRaises(IndexError, self.a.pop)
        i = 3
        while len(self.b) > 0:
            v = self.b.pop()
            self.assertEqual(v, i)
            i -= 1

    def test_top(self):
        self.assertRaises(IndexError, self.a.top)
        self.assertEqual(self.b.top(), 3)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Version: 6.0 (11-dic-2024)
         5.2 (01-dic-2023)
         5.1 (23-nov-2023)
         5.0 (31-oct-2023)
         4.1 (29-sep-2022)
         4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2023
@license: GPL3
"""
import operator
from abc import abstractmethod
from functools import total_ordering
from typing import final, Self, Callable

from algoritmia.datastructures.priorityqueues import MaxHeap, MinHeap, IPriorityQueue
from algoritmia.schemes.bt_scheme import DecisionSequence, DecisionPath


# ACERCA DEL TIPO ScoredSolution (importado de bt_scheme)
# Las funciones bab_min_solve y bab_max_solve devuelven Optional[ScoredSolution]:
#   - Devuelven 'None' si no hay solución.
#   - Devuelven la tupla '(score, ds_sol)' si hay solución: donde 'ds_sol' es la
#     BabDecisionSequence que lleva a la solución óptima y 'score', su puntuación.

# ACERCA DEL CÁLCULO EFICIENTE DE LAS COTAS
# Dado que las cotas son inmutables solo deberían calcularse una vez:
# - Los métodos calculate_opt_bound() y calculate_pes_bound() se llaman sólo en el constructor
#   y sus resultados se guardan en los atributos self._opt y self._pes.
# - Los métodos opt() y pes() sólo devuelven el valor de estos atributos.

# La clase BabDecisionSequence -------------------------------------------------------


@total_ordering  # Implementando < y ==, el resto de operadores de comparación se generan automáticamente
class BabDecisionSequence[TDecision, TExtra, TScore](DecisionSequence[TDecision, TExtra]):
    def __init__(self,
                 extra: TExtra | None = None,
                 decisions: DecisionPath[TDecision] = (),
                 length: int = 0):
        DecisionSequence.__init__(self, extra, decisions, length)
        self._pes = self.calculate_pes_bound()
        self._opt = self.calculate_opt_bound()

    # --- Métodos abstractos nuevos ---

    @abstractmethod
    def calculate_opt_bound(self) -> TScore:  # Calcula y devuelve la cota optimista
        pass

    @abstractmethod
    def calculate_pes_bound(self) -> TScore:  # Calcula y devuelve la cota pesimista
        pass

    # --- Métodos abstractos heredados  ---

    # @abstractmethod
    # def successors(self) -> Iterator[Self]:
    #     pass

    # @abstractmethod
    # def is_solution(self) -> bool:
    #     pass

    # --- Método heredado que puede sobreescribirse en las clases hijas ---

    # def state(self) -> State:     # La implementación por defecto es O(n)

    # -- Métodos finales que NO se pueden sobreescribir en las clases hijas ---

    # Cota optimista. Para las soluciones su valor debe coincidir con la puntuación real
    @final
    def opt(self) -> TScore:
        return self._opt

    # Cota pesimista. Para las soluciones su valor debe coincidir con la puntuación real
    @final
    def pes(self) -> TScore:
        return self._pes

    # Comparar dos BabDecisionSequence es comparar sus cotas optimistas
    @final
    def __lt__(self, other: Self) -> bool:
        return self._opt < other._opt

    @final
    def __eq__(self, other: Self) -> bool:
        return self._opt == other._opt

    # -- Métodos finales heredados que NO pueden sobreescribirse en las clases hijas ---

    # @final
    # def add_decision(self, decision: TDecision, extra: TExtra = None) -> Self:    # O(1)

    # @final
    # def decisions(self) -> tuple[TDecision, ...]:     # O(n)

    # @final
    # def last_decision(self) -> TDecision:             # O(1)

    # @final
    # def __len__(self) -> int:                         # O(1)

# Esquemas para BaB --------------------------------------------------------------------------

def bab_solve[TDecision, TExtra, TScore](better: Callable[[TScore, TScore], bool],
                                         heap: IPriorityQueue[TScore],
                                         initial_ds: BabDecisionSequence[TDecision, TExtra, TScore]
                                         ) -> tuple[TScore, BabDecisionSequence[TDecision, TExtra, TScore]] | None:
    bps = initial_ds.pes()
    heap.add(initial_ds)
    best_seen = {initial_ds.state(): initial_ds.opt()}
    while len(heap) > 0:
        best_ds = heap.extract_opt()
        if best_ds.is_solution():
            return best_ds.opt(), best_ds
        for new_ds in best_ds.successors():
            new_opt = new_ds.opt()
            if not better(bps, new_opt):
                if better(new_ds.pes(), bps):
                    bps = new_ds.pes()
                new_state = new_ds.state()
                bs = best_seen.get(new_state, None)
                if bs is None or better(new_opt, bs):
                    best_seen[new_state] = new_opt
                    heap.add(new_ds)


def bab_min_solve[TDecision, TExtra, TScore](initial_ds: BabDecisionSequence[TDecision, TExtra, TScore])\
        -> tuple[TScore, BabDecisionSequence[TDecision, TExtra, TScore]] | None:
    return bab_solve(operator.lt, MinHeap(), initial_ds)


def bab_max_solve[TDecision, TExtra, TScore](initial_ds: BabDecisionSequence[TDecision, TExtra, TScore])\
        -> tuple[TScore, BabDecisionSequence[TDecision, TExtra, TScore]] | None:
    return bab_solve(operator.gt, MaxHeap(), initial_ds)


from abc import abstractmethod, ABC
from collections.abc import Callable, Iterator, Sequence, Sized
from itertools import chain, repeat


class IPriorityQueue[T](ABC, Sized):
    @abstractmethod
    def add(self, item: T): pass

    @abstractmethod
    def opt(self) -> T: pass

    @abstractmethod
    def extract_opt(self) -> T: pass


class MinHeap[T](IPriorityQueue[T]):
    _opt: Callable[[T, T], T] = min

    def __init__(self, data: Sequence[T] = (), capacity: int = 0):
        # self._opt = self.opt
        self._size = len(data)
        self._heap: list[T] = list(chain((None,), data, repeat(None, max(0, capacity - self._size))))
        for i in range(self._size // 2, 0, -1):
            self._heapify(i)

    def _heapify(self, i: int):
        while True:
            left: int = 2 * i
            right: int = left + 1
            if left <= self._size and self._opt(self._heap[left], self._heap[i]) != self._heap[i]:
                best = left
            else:
                best = i
            if right <= self._size and self._opt(self._heap[right], self._heap[best]) != self._heap[best]:
                best = right
            if best == i:
                break
            self._heap[i], self._heap[best] = self._heap[best], self._heap[i]
            i = best

    def opt(self) -> T:
        if self._size == 0:
            raise IndexError('opt from an empty heap')
        return self._heap[1]

    def extract_opt(self) -> T:
        if self._size == 0:
            raise IndexError('extract opt from an empty heap')
        m = self._heap[1]
        if self._size > 1:
            self._heap[1], self._heap[self._size] = self._heap[self._size], None
        self._size -= 1
        if self._size > 1:
            self._heapify(1)
        return m

    def add(self, item: T):
        if self._size + 1 == len(self._heap):
            self._heap.append(None)
        i = self._size = self._size + 1
        self._heap[i] = item
        self._bubble_up(i)

    def _bubble_up(self, i: int):
        parent = i // 2
        while i > 1 and self._opt(self._heap[i], self._heap[parent]) != self._heap[parent]:
            self._heap[i], self._heap[parent] = self._heap[parent], self._heap[i]
            i, parent = parent, parent // 2

    def __iter__(self) -> Iterator[T]:
        for i in range(1, self._size + 1): yield self._heap[i]

    def __len__(self) -> int:
        return self._size

    def __repr__(self) -> str:
        if self.__class__.__name__ == 'Heap':
            b = 'min' if self._opt == min else 'max'
            return 'Heap({}, {!r})'.format(b, self._heap[1:self._size + 1])
        return '{}({!r})'.format(self.__class__.__name__, self._heap[1:self._size + 1])


class MaxHeap[T](MinHeap[T]):
    _opt = max



from algoritmia.algorithms.traverse import traverse_dijkstra_dict, traverse_dijkstra_metric_dict
from algoritmia.data.iberia import iberia, km, coords2d, iberia2d
from algoritmia.viewers.graph2d_viewer import Graph2dViewer


def dist(city_a: str, city_b: str) -> float:
    coords_a, coords_b = coords2d[city_a], coords2d[city_b]
    dx, dy = coords_a[0] - coords_b[0], coords_a[1] - coords_b[1]
    return (dx * dx + dy * dy) ** 0.5


if __name__ == '__main__':
    v_initial = 'Madrid'
    v_final = 'Bilbao'
    edges = traverse_dijkstra_dict(iberia, km, v_initial)
    # edges = traverse_dijkstra_metric_dict(iberia, km, dist,v_initial, v_final)
    colors = {}
    for (c1, c2) in edges:
        c1_pos = coords2d[c1]
        c2_pos = coords2d[c2]
        colors[c2_pos] = 'red'
        if c2 == v_final: break
    colors[coords2d[v_initial]] = colors[coords2d[v_final]] = 'palegreen'

    gv = Graph2dViewer(iberia2d, canvas_width=800, canvas_height=800, vertexmode=Graph2dViewer.X_Y, colors=colors)
    gv.run()



from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction

type City = str

_km = WeightingFunction[City](
    {('Alcúdia', 'Artà'): 36, ('Alcúdia', 'Inca'): 25,
     ('Alcúdia', 'Pollença'): 10, ('Andratx', 'Calvià'): 14,
     ('Andratx', 'Palma de Mallorca'): 30, ('Andratx', 'Sóller'): 56,
     ('Artà', 'Capdepera'): 8, ('Artà', 'Manacor'): 17,
     ('Calvià', 'Palma de Mallorca'): 14, ('Campos del Port', 'Llucmajor'): 14,
     ('Campos del Port', 'Santanyí'): 13, ('Inca', 'Manacor'): 25,
     ('Inca', 'Marratxí'): 12, ('Llucmajor', 'Palma de Mallorca'): 20,
     ('Manacor', 'Santanyí'): 27, ('Marratxí', 'Palma de Mallorca'): 14,
     ('Pollença', 'Sóller'): 54},
    symmetrical=True)

Mallorca = UndirectedGraph[City](E=_km.keys())



from collections.abc import Iterator
from dataclasses import dataclass
from typing import Self

from math import ceil

from algoritmia.schemes.bab_scheme import BabDecisionSequence, bab_min_solve
from algoritmia.utils import infinity

# Tipos  --------------------------------------------------------------------------

type Decision = int  # Número de monedas del tipo actual
type Solution = tuple[Decision, ...]

# 'bab_min_solve' devuelve Optional[ScoredSolution]
type Score = int  # Total de monedas utilizado
type ScoredSolution = tuple[Score, Solution]


# --------------------------------------------------------------------------------

def coin_change_bab_solve(v: tuple[int, ...], Q: int) -> ScoredSolution | None:
    @dataclass
    class Extra:
        pending: int
        used_coins: int

    class CoinChangeDS(BabDecisionSequence[Decision, Extra, Score]):
        def calculate_opt_bound(self) -> Score:
            s = self.extra.used_coins
            if self.extra.pending > 0:
                if len(self) == len(v):
                    return infinity
                s += ceil(self.extra.pending / max(v[len(self):]))
            return s

        def calculate_pes_bound(self) -> Score:
            if self.is_solution():
                return self.extra.used_coins
            return infinity  # Mejorable

        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    new_pending = self.extra.pending - num_coins * v[n]
                    new_used_coins = self.extra.used_coins + num_coins
                    yield self.add_decision(num_coins, Extra(new_pending, new_used_coins))

        def state(self) -> tuple[int, int]:
            return len(self), self.extra.pending

    initial_ds = CoinChangeDS(Extra(Q, 0))
    result = bab_min_solve(initial_ds)
    if result is None: return None
    score, solution_ds = result
    return score, solution_ds.decisions()


# Programa principal --------------------------------------------------------------------------------

if __name__ == "__main__":
    v0, Q0 = (1, 2, 5), 7
    print(f"Coins: {v0}")
    print(f"Quantity: {Q0}\n")

    sol0 = coin_change_bab_solve(v0, Q0)
    if sol0 is not None:
        print(f"Best coin change solution: {sol0}")
    else:
        print("There are no solutions")



from typing import Optional


def binarysearch_tail_rec(v: list[int], elem: int) -> Optional[int]:
    def rec(start: int, end: int):
        if end - start == 0:
            return None
        mid = (start + end) // 2
        if elem < v[mid]:
            return rec(start, mid)
        elif elem > v[mid]:
            return rec(mid + 1, end)
        else:  # elem == v[mid]:
            return mid

    return rec(0, len(v))


def binarysearch_iter(v: list[int], elem: int) -> Optional[int]:
    start, end = 0, len(v)
    while end - start > 0:
        mid = (start + end) // 2
        if elem < v[mid]:
            end = mid
        elif elem > v[mid]:
            start = mid + 1
        else:  # elem == v[mid]:
            return mid


if __name__ == "__main__":
    my_sorted_list = [2, 3, 3, 11, 12, 18, 21, 29, 30, 37, 43, 75, 82, 98]
    num = 30
    print('list:', my_sorted_list)
    print('number to find:', num)

    pos = binarysearch_tail_rec(my_sorted_list, num)
    print(f'binarysearch_tail_rec: number found at index {pos}')

    pos = binarysearch_iter(my_sorted_list, num)
    print(f'binarysearch_iter: number found at index {pos}')



from algoritmia.datastructures.graphs import UndirectedGraph

if __name__ == "__main__":
    edges = [('Castelló', 'Sagunt'), ('Sagunt', 'València')]

    ugraph = UndirectedGraph(E=edges)  # Parameter V is optional
    print(ugraph.V)  # Output: {'Sagunt', 'València', 'Castelló'}
    print(ugraph.succs('Sagunt'))  # Output: {'València', 'Castelló'}



import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import Optional

# T is a generic with the vertex type
type Edge[T] = tuple[T, T]  # Edge is a generic type
type Weight = int | float  # Edge weight

# Crea grafos con las siguientes características:
# 1. Una arista como máximo entre vértices:
#    - Si el grafo es dirigido, (u, v) y (v, u) se consideran aristas distintas.
#    - Si el grafo es no dirigido, (u, v) y (v, u) se consideran la misma arista y solo se almacena la primera
#      en la lista de aristas.
# 2. Se descartan las auto aristas (v, v)


# UndirectedGraph, Digraph -----------------------------------------------------


# Abstract class (no instances)
class IGraph[T](ABC):
    @abstractmethod
    def is_directed(self) -> bool:
        pass

    @property
    def V(self) -> set[T]:
        return self._v

    @property
    def E(self) -> list[Edge[T]]:
        return self._e

    # O(|V| + |E|)
    def __init__(self, V: Optional[Iterable[T]] = None, E: Iterable[Edge[T]] = ()):
        # Crea el conjunto de vértices self._v
        if V is None:
            self._v: set[T] = set(v for e in E for v in e)
        else:
            self._v: set[T] = set(V)

        # Crea la lista de aristas self._e y los diccionarios self._s y self._p
        self._e: list[Edge[T]] = []
        self._s: dict[T, set[T]] = dict((v, set()) for v in self._v)
        self._p: dict[T, set[T]] = dict((v, set()) for v in self._v) if self.is_directed() else self._s
        for e in E:
            u, v = e
            if u not in self._v:
                raise TypeError(f"{self.__class__.__name__} - Vertex {u} from edge {e} is not in the vertex set")
            if v not in self._v:
                raise TypeError(f"{self.__class__.__name__} - Vertex {v} from edge {e} is not in the vertex set")
            if u == v:
                sys.stderr.write(f"{self.__class__.__name__} - WARNING: Discarded auto edge {e}\n")
                continue
            if v in self._s[u]:
                sys.stderr.write(f"{self.__class__.__name__} - WARNING: Discarded repeated edge {e}\n")
                continue

            self._s[u].add(v)
            self._p[v].add(u)
            self._e.append(e)

    # O(1)*
    def succs(self, v: T) -> set[T]:
        return self._s[v]

    # O(1)*
    def preds(self, v: T) -> set[T]:
        return self._p[v]

    # O(1)*
    def out_degree(self, u: T) -> int:
        return len(self._s[u])

    # O(1)*
    def in_degree(self, v: T) -> int:
        return len(self._p[v])

    # O(1)*
    def add_vertex(self, v: T):
        if v in self._v: return
        assert v not in self._s, 'add_vertex - Impossible 1'
        assert v not in self._p, 'add_vertex - Impossible 2'
        self._v.add(v)
        self._s[v] = set()
        self._p[v] = set()

    # O(|V| + |E|)
    def remove_vertex(self, v: T):
        if v not in self._v: return
        assert v in self._s, 'remove_vertex - Impossible 3'
        assert v in self._p, 'remove_vertex - Impossible 4'
        self._v.remove(v)

        for suc in self._s[v]:
            self._p[suc].remove(v)  # O(1)*, |V| veces: O(|V|)
        del self._s[v]
        if self.is_directed():
            for pred in self._p[v]:
                self._s[pred].remove(v)  # O(1)*, |V| veces: O(|V|)
            del self._p[v]

        self._e = [(u, w) for (u, w) in self._e if u != v and w != v]  # O(|E|)

    # O(1)*
    def add_edge(self, e: Edge[T]):
        u, v = e
        self.add_vertex(u)
        self.add_vertex(v)

        self._e.append(e)

        self._s[u].add(v)
        self._p[v].add(u)

    # O(|E|)
    def remove_edge(self, e: Edge[T]):
        u, v = e
        if u in self._s: self._s[u].discard(v)
        if v in self._p: self._p[v].discard(u)
        if e in self._e: self._e.remove(e)  # O(|E|)
        if not self.is_directed() and (v, u) in self._e:  # O(|E|)
            self._e.remove((v, u))  # O(|E|)

    # O(1)*
    def contains_vertex(self, v: T) -> bool:
        return v in self._v

    # O(1)*
    def contains_edge(self, e: Edge[T]) -> bool:
        u, v = e
        if self.is_directed():
            return u in self._s and v in self._s[u]
        return u in self._s and v in self._s[u] or v in self._s and u in self._s[v]

    # O(|V| + |E|)
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(V={list(self.V)}, E={list(self.E)})"


class UndirectedGraph[T](IGraph[T]):
    # O(1)
    def is_directed(self) -> bool:
        return False


class Digraph[T](IGraph[T]):
    # O(1)
    def is_directed(self) -> bool:
        return True


# WeightingFunction -----------------------------------------------------------------


class WeightingFunction[T](dict[Edge[T], Weight], Callable[[T | Edge[T], Optional[T]], Weight]):
    # O(|data|) = O(|E|)
    def __init__(self, data: Iterable[tuple[Edge[T], Weight]] | dict[Edge[T], Weight], symmetrical: bool = False):
        super().__init__(data)
        self.symmetrical = symmetrical
        to_delete: list[Edge[T]] = []
        edges: set[Edge[T]] = set()
        for e in self.keys():
            u, v = e
            if u == v:
                sys.stderr.write(
                    f"WeightingFunction - WARNING: Discarded auto edge {e} (with weight {self[e]}).\n")
                continue
            if symmetrical:
                e2 = v, u
                if e2 in edges:
                    if self[e2] != self[e]:
                        m = f"Weights are symmetrical but {e2} and {e} have differents weights: {self[e2]} != {self[e]}"
                        raise ValueError(m)
                    sys.stderr.write(
                        f"WeightingFunction - WARNING: Discarded repeated edge {e} (with weight {self[e]}).\n")
                    to_delete.append(e)  # dejar solo una arista en grafos no dirigidos
                else:
                    edges.add(e)
        for e in to_delete:
            del self[e]

    # O(1)*
    def __call__(self, u: T | Edge[T], v: Optional[T] = None) -> Weight:
        if v is None:
            u, v = u
        # if u == v:
        #    return 0
        if (u, v) in self:
            return self[u, v]
        if self.symmetrical:
            if (v, u) in self:
                return self[v, u]
            raise KeyError(repr((u, v)) + " nor " + repr((v, u)))
        raise KeyError(repr((u, v)))


if __name__ == '__main__':
    g = Digraph(E=[(1, 2)])
    print(g)
    print(isinstance(g, IGraph))
    print(isinstance(g, Digraph))
    print(isinstance(g, UndirectedGraph))



# coding: latin1
import unittest

from algoritmia.utils import argmax, argmin, infinity


class TestMinMax(unittest.TestCase):
    def testInfinty(self):
        self.assertTrue(infinity > 1)
        self.assertTrue(infinity > 2 ** 1000)
        self.assertTrue(-infinity < 2 ** 1000)

    def test_argmin(self):
        self.assertEqual(argmin((1, 2, 3, 4), lambda x: -x), 4)
        self.assertEqual(argmin([], lambda x: -x), None)

    def test_argmax(self):
        self.assertEqual(argmax((1, 2, 3, 4), lambda x: -x), 1)
        self.assertEqual(argmax([], lambda x: -x), None)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# coding: latin1

import unittest

from algoritmia.datastructures.graphs import (UndirectedGraph, Digraph,
                                              WeightingFunction)


class TestDigraphs(unittest.TestCase):
    def setUp(self):
        self.edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0), (0, 3), (3, 0), (0, 2), (0, 4), (0, 7)]
        self.g1 = Digraph(V=range(8), E=self.edges)
        self.g1bis = Digraph(E=self.edges)
        self.g1bis.add_vertex(6)
        edges = self.edges + [(1, 3)]
        self.g6 = UndirectedGraph(V=range(8), E=self.edges)
        self.g8 = Digraph(V=range(8), E=edges)

    def test_readonly(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(len(g.V), 8)
            self.assertEqual(len(g.E), len(self.edges))
            for (u, v) in self.edges:
                self.assertTrue((u, v) in g.E)
            for u in g.V:
                succ = set(v for (uu, v) in self.edges if uu == u)
                pred = set(v for (v, uu) in self.edges if uu == u)
                self.assertEqual(succ, set(g.succs(u)))
                self.assertEqual(pred, set(g.preds(u)))
            self.assertFalse((0, 0) in g.E)

            self.assertEqual(g.out_degree(0), 5)
            self.assertEqual(g.in_degree(0), 2)

    def test_remove_self_loop(self):
        self.g8.remove_vertex(1)
        self.assertEqual(9, len(self.g8.E))

    def test_edit_edges(self):
        for g in self.g1, self.g1bis:
            self.assertFalse(1 in g.succs(7))
            self.assertFalse(7 in g.succs(1))
            self.assertFalse(7 in g.preds(1))
            self.assertFalse(1 in g.preds(7))
            g.add_edge((7, 1))
            self.assertEqual(len(g.E), len(self.edges) + 1)
            self.assertTrue((7, 1) in g.E)
            self.assertFalse((1, 7) in g.E)
            self.assertTrue(1 in g.succs(7))
            self.assertFalse(7 in g.succs(1))
            self.assertTrue(7 in g.preds(1))
            self.assertFalse(1 in g.preds(7))

            self.assertTrue(7 in g.succs(0))
            self.assertFalse(0 in g.succs(7))
            self.assertTrue(0 in g.preds(7))
            self.assertFalse(7 in g.preds(0))
            g.remove_edge((0, 7))
            self.assertFalse((0, 7) in g.E)
            self.assertEqual(len(g.E), len(self.edges))
            self.assertFalse(0 in g.succs(7))
            self.assertFalse(7 in g.succs(0))
            self.assertFalse(7 in g.preds(0))
            self.assertFalse(0 in g.preds(7))

    def test_edit_vertices(self):
        for g in self.g1, self.g1bis:
            g.add_vertex(6)
            self.assertTrue(6 in g.V)
            self.assertFalse((0, 6) in g.E)
            self.assertFalse(0 in g.succs(6))
            self.assertFalse((6, 0) in g.E)
            self.assertFalse(6 in g.succs(0))

            if g == self.g6:  # Por cobertura
                g.add_vertex(10)

            g.remove_vertex(0)
            self.assertFalse((1, 0) in g.E)
            self.assertFalse((0, 1) in g.E)
            self.assertFalse(0 in set(s for v in g.V for s in g.succs(v)))
            self.assertFalse(0 in set(s for v in g.V for s in g.preds(v)))

    def test_contains(self):
        for g in self.g1, self.g1bis:
            self.assertFalse(g.contains_edge((7, 1)))
            self.assertFalse(g.contains_edge((1, 7)))
            g.add_edge((7, 1))
            self.assertTrue(g.contains_edge((7, 1)))
            self.assertFalse(g.contains_edge((1, 7)))

class TestUndirectedGraph(unittest.TestCase):
    def setUp(self):
        self.edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0), (0, 3), (0, 2), (0, 4), (0, 7)]
        self.undirected_edges = set()
        for (u, v) in self.edges:
            self.undirected_edges.add((u, v))
            self.undirected_edges.add((v, u))
        self.g1 = UndirectedGraph(V=range(8), E=self.edges)
        self.g1bis = UndirectedGraph(E=self.edges)
        self.g1bis.add_vertex(6)
        self.g6 = UndirectedGraph(V=range(8), E=self.edges)
        self.g7 = UndirectedGraph(V=range(8), E=self.edges)

    def test_V_contains(self):
        for g in self.g1, self.g1bis:
            for i in range(8):
                self.assertTrue(i in g.V)
                self.assertFalse(8 + i in g.V)

    def test_E_contains(self):
        for g in self.g1, self.g1bis:
            for (u, v) in self.edges:
                self.assertTrue((u, v) in g.E or (v, u) in g.E)
                self.assertFalse((u, v + 8) in g.E)
                self.assertFalse((u + 8, v) in g.E)

    def test_remove_vertex(self):
        self.g7.remove_vertex(0)
        self.assertEqual(4, len(self.g7.E))

    def test_V_len(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(len(g.V), 8)

    def test_E_len(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(len(g.E), 10)

    def test_V_repr(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(set(g.V), set(eval(repr(g.V))))

    def test_E_repr(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(set(g.E), set(eval(repr(g.E))))

    def test_succs_and_preds(self):
        for g in self.g1, self.g1bis:
            for u in g.V:
                succ = set(v for (uu, v) in self.undirected_edges if uu == u)
                pred = set(v for (v, uu) in self.undirected_edges if uu == u)
                self.assertEqual(succ, set(g.succs(u)))
                self.assertEqual(pred, set(g.preds(u)))

    def test_out_degree(self):
        for g in self.g1, self.g1bis:
            self.assertEqual(g.out_degree(0), 6)
            for v in g.V:
                self.assertEqual(g.out_degree(v), g.in_degree(v))

    def test_repr(self):
        for g in self.g1, self.g1bis:
            gg = eval(repr(g))
            self.assertEqual(set(g.V), set(gg.V))
            self.assertEqual(set(g.E), set(gg.E))

    def test_edit_edges(self):
        for g in self.g1, self.g1bis:
            g.add_edge((7, 1))
            # self.assertEqual(len(g.E), len(self.undirected_edges)+2)
            self.assertTrue((7, 1) in g.E)
            # self.assertTrue((1,7) in g.E)
            self.assertTrue(7 in g.succs(1))
            self.assertTrue(1 in g.succs(7))
            self.assertTrue(7 in g.preds(1))
            self.assertTrue(1 in g.preds(7))


            g.remove_edge((0, 7))
            # self.assertFalse((7,0) in g.E)
            self.assertFalse((0, 7) in g.E)
            self.assertFalse(7 in g.succs(0))
            self.assertFalse(0 in g.succs(7))
            self.assertFalse(7 in g.preds(0))
            self.assertFalse(0 in g.preds(7))

            g.add_edge((1, 3))
            self.assertTrue((1, 3) in g.E)
            # self.assertTrue((3,1) in g.E)
            self.assertTrue(1 in g.succs(3))
            self.assertTrue(3 in g.succs(1))
            self.assertTrue(1 in g.preds(3))
            self.assertTrue(3 in g.preds(1))

    def test_edit_vertices(self):
        for g in self.g1, self.g1bis:
            g.add_vertex(6)
            self.assertTrue(6 in g.V)
            self.assertFalse((0, 6) in g.E)
            self.assertFalse(0 in g.succs(6))
            self.assertFalse((6, 0) in g.E)
            self.assertFalse(6 in g.succs(0))

            if g == self.g6:  # Por cobertura
                g.add_vertex(10)

            g.remove_vertex(0)
            self.assertFalse((1, 0) in g.E)
            self.assertFalse((0, 1) in g.E)
            self.assertFalse(0 in set(s for v in g.V for s in g.succs(v)))
            self.assertFalse(0 in set(s for v in g.V for s in g.preds(v)))

    def test_contains(self):
        for g in self.g1, self.g1bis:
            self.assertFalse(g.contains_edge((7, 1)))
            self.assertFalse(g.contains_edge((1, 7)))
            g.add_edge((7, 1))
            self.assertTrue(g.contains_edge((7, 1)))
            self.assertTrue(g.contains_edge((1, 7)))

class TestRedimDigraphs(unittest.TestCase):
    def setUp(self):
        self.edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0), (0, 3), (3, 0), (0, 2), (0, 4), (0, 7)]
        self.g = Digraph(V=range(8), E=self.edges)

    def test_edit_edges(self):
        for g in [self.g]:
            g.add_edge((7, 1))
            self.assertEqual(len(g.E), len(self.edges) + 1)
            self.assertTrue((7, 1) in g.E)
            self.assertFalse((1, 7) in g.E)

            g.remove_edge((0, 7))
            self.assertFalse((7, 0) in g.E)
            self.assertFalse((0, 7) in g.E)

            g.add_edge((12, 1))
            self.assertTrue((12, 1) in g.E)
            self.assertFalse((1, 12) in g.E)

            g.add_edge((1, 12))
            self.assertTrue((1, 12) in g.E)


class TestWeightingFunction(unittest.TestCase):
    def setUp(self):
        self.sym = WeightingFunction([((0, 0), 1), ((1, 1), 2), ((1, 2), 3)], symmetrical=True)
        self.sym_d = WeightingFunction({(0, 0): 1, (1, 1): 2, (1, 2): 3}, symmetrical=True)
        self.asym = WeightingFunction([((0, 0), 1), ((1, 1), 2), ((1, 2), 3)], symmetrical=False)
        self.asym_d = WeightingFunction({(0, 0): 1, (1, 1): 2, (1, 2): 3}, symmetrical=False)
        self.rep = WeightingFunction([((0, 1), 10), ((0, 2), 2), ((1, 0), 10)], symmetrical=True)

    def test_getitem(self):
        for wf in [self.sym, self.sym_d]:
            self.assertEqual(wf(0, 0), 1)
            self.assertEqual(wf(1, 1), 2)
            self.assertEqual(wf(1, 2), 3)
            self.assertEqual(wf(2, 1), 3)
        for wf in [self.asym, self.asym_d]:
            self.assertEqual(wf(0, 0), 1)
            self.assertEqual(wf(1, 1), 2)
            self.assertEqual(wf(1, 2), 3)
            self.assertRaises(KeyError, wf.__call__, (2, 1))
            self.assertRaises(KeyError, wf.__call__, (2, 1))

        self.assertRaises(ValueError, WeightingFunction, *[[((0, 1), 1), ((0, 2), 2), ((1, 0), 3)], True])
        self.assertEqual(self.rep(0, 1), 10)
        self.assertEqual(self.rep(1, 0), 10)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from collections.abc import Iterator
from dataclasses import dataclass
from random import random, seed
from typing import Self

from algoritmia.schemes.bab_scheme import BabDecisionSequence, bab_min_solve
from algoritmia.utils import infinity

# Tipos  --------------------------------------------------------------------------

Decision = int
Solution = tuple[Decision, ...]

# 'bab_min_solve' devuelve Optional[ScoredSolution]
Score = int
ScoredSolution = tuple[Score, Solution]


# --------------------------------------------------------------------------------

def sumset_bab_solve(e: tuple[int, ...], s: int) -> ScoredSolution | None:
    @dataclass
    class Extra:
        acc_sum: int = 0  # accumulated sum
        used_nums: int = 0

    class SumSetDS(BabDecisionSequence[Decision, Extra, Score]):
        def calculate_opt_bound(self) -> Score:
            if self.extra.acc_sum == s:
                return self.extra.used_nums
            return self.extra.used_nums + 1  # Mejorable

        def calculate_pes_bound(self) -> Score:
            if self.extra.acc_sum == s:
                return self.extra.used_nums
            return infinity  # Mejorable

        def is_solution(self) -> bool:
            return len(self) == len(e) and self.extra.acc_sum == s

        def successors(self) -> Iterator[Self]:
            if len(self) < len(e):
                yield self.add_decision(0, self.extra)
                new_acc = self.extra.acc_sum + e[len(self)]
                if new_acc <= s:
                    new_used_nums = self.extra.used_nums + 1
                    yield self.add_decision(1, Extra(new_acc, new_used_nums))

        # Sobreescribimos 'state()'
        def state(self) -> tuple[int, int]:
            return len(self), self.extra.acc_sum

    initial_ds = SumSetDS(Extra())
    result = bab_min_solve(initial_ds)
    if result is None: return None
    score, solution_ds = result
    return score, solution_ds.decisions()


# --------------------------------------------------------------------------------

def subsetsum_problem(num_elem):
    seed(42)
    elems = tuple(int(random() * 1000) + 1 for _ in range(num_elem))
    return elems, sum(elems) // 4  # sorted(elems, key=lambda x: -x), sum(elems) // 4


# --------------------------------------------------------------------------------

if __name__ == "__main__":
    # Instance
    elements0, target0 = (640, 777, 276, 224, 677, 893, 87, 422, 30, 306), 393
    print(f"Elements: {elements0}")
    print(f"Target: {target0}\n")

    sol0 = sumset_bab_solve(elements0, target0)
    if sol0 is not None:
        print(f"Best coin change solution: {sol0}")
    else:
        print("There are no solutions")



from collections.abc import Iterator

from algoritmia.algorithms.traverse import Traverse
from algoritmia.datastructures.graphs import UndirectedGraph


def spanning_forest[T](g: UndirectedGraph[T], traverser: Traverse[T]) -> Iterator[UndirectedGraph]:
    pending_vertices = set(g.V)
    while len(pending_vertices) > 0:
        u = pending_vertices.pop()
        e_gen = iter(traverser(g, u))
        next(e_gen)  # skip phantom edge
        visited_edges = list(e_gen)
        pending_vertices -= set([v for (u, v) in visited_edges])
        yield UndirectedGraph(E=list(visited_edges))


if __name__ == '__main__':
    from traverse import traverse_bf, traverse_df

    type Vertex = tuple[int, int]

    edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((2, 0), (2, 1)),
             ((2, 2), (2, 3)), ((0, 1), (1, 1)), ((0, 2), (1, 2)), ((0, 3), (1, 3)),
             ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=edges)

    print("Breath first:")
    graphs_bf = list(spanning_forest(my_graph, traverse_bf))
    for g0 in graphs_bf:
        print(g0)

    print("Depth first:")
    graphs_df = list(spanning_forest(my_graph, traverse_df))
    for g0 in graphs_df:
        print(g0)



from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from typing import Optional


# Lista enlazada con puntero a cabeza y cola.
# Cumple la interfaz list de Python.
# Costes: l.append(e), l.pop(), l.insert(0, e), del l[0], len(l) => O(1)
#         l[index], l.remove(elem), l.reverse(), elem in l => O(n)


class LinkedList[T](list[T]):
    @dataclass
    class Node[T]:
        value: T
        prev: Optional['LinkedList.Node[T]']
        next: Optional['LinkedList.Node[T]']

    def __init__(self, seq: Iterable[T] = ()):
        super().__init__()
        self._head: Optional[LinkedList.Node[T]] = None
        self._last: Optional[LinkedList.Node[T]] = None
        self._length: int = 0
        for item in seq:
            self.append(item)

    def append(self, value: T):
        node = LinkedList.Node(value, self._last, None)
        if self._last is None:
            self._head = self._last = node
        else:
            self._last.next = node
            self._last = node
        self._length += 1

    def _get_node_at(self, i: int) -> Node[T]:
        if not (-self._length <= i < self._length): raise IndexError("{}".format(i))
        if i > self._length // 2:
            i -= self._length
        elif i < -self._length // 2:
            i += self._length
        if i >= 0:
            node, j = self._head, 0
            while node and j < i: node, j = node.next, j + 1
        else:
            i, node, j = -i, self._last, 1
            while node and j < i: node, j = node.prev, j + 1
        return node

    def __getitem__(self, i: int) -> T:
        node = self._get_node_at(i)
        if node is None: raise IndexError('No item at position {!r}'.format(i))
        return node.value

    def __setitem__(self, i: int, value: T) -> T:
        node = self._get_node_at(i)
        if node is None: raise IndexError('No item at position {!r}'.format(i))
        node.value = value
        return value

    def _remove(self, node: Node[T]):
        if node.prev: node.prev.next = node.next
        if node.next: node.next.prev = node.prev
        if self._head == node: self._head = node.next
        if self._last == node: self._last = node.prev

    def __delitem__(self, i: int):
        node = self._get_node_at(i)
        if node is None: raise IndexError('No item at position {!r}'.format(i))
        self._remove(node)
        self._length -= 1

    def remove(self, value: T):
        if self._head is not None:
            node = self._head
            while node.value != value and node.next: node = node.next
            if node and node.value == value:
                self._remove(node)
                self._length -= 1
                return
        raise ValueError("Cannot remove {!r}".format(value))

    def pop(self, i: int = None) -> T:
        if i is None:
            if self._last is None: raise IndexError('pop from an empty linked list')
            value = self._last.value
            self._remove(self._last)
            self._length -= 1
        else:
            node = self._get_node_at(i)
            if node is None: raise IndexError('No item at position {!r}'.format(i))
            value = node.value
            self._remove(node)
            self._length -= 1
        return value

    def insert(self, i: int, item: T):
        if i == 0:
            newnode = LinkedList.Node(item, None, self._head)
            if self._head is not None:
                self._head.prev = newnode
                self._head = newnode
            else:
                self._head = self._last = newnode
            self._length += 1
        elif i == self._length:
            self.append(item)
        else:
            node = self._get_node_at(i - 1)
            if node is None: raise IndexError('Cannot insert at position {!r}'.format(i))
            newnode = LinkedList.Node(item, node, node.next)
            if node.next: node.next.prev = newnode
            node.next = newnode
            if node == self._last: self._last = newnode
            self._length += 1

    def reverse(self):
        left, right = self._head, self._last
        for _ in range(len(self) // 2):
            left.value, right.value = right.value, left.value
            left = left.next
            right = right.prev

    def __iter__(self) -> Iterator[T]:
        node = self._head
        while node is not None:
            yield node.value
            node = node.next

    def __contains__(self, value: T) -> bool:
        return any(value == v for v in self)

    def __len__(self) -> int:
        return self._length

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, list(self))



def convex_min(v: list[int]) -> int:
    def rec(start: int, end: int) -> int:
        if end - start == 1:
            return v[start]
        elif end - start == 2:
            return min(v[start], v[start + 1])
        else:
            mid = (start + end) // 2
            if v[mid - 1] < v[mid]:
                return rec(start, mid)
            else:
                return rec(mid, end)

    return rec(0, len(v))


if __name__ == '__main__':
    v0 = [9, 8, 6, 5, 4, 4, 3, 2, 5, 6, 8]
    print(convex_min(v0))



from collections.abc import Iterator
from dataclasses import dataclass
from typing import Self

from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction
from algoritmia.schemes.bab_scheme import BabDecisionSequence, bab_min_solve
from algoritmia.utils import infinity

# Tipos  --------------------------------------------------------------------------

# En este problema, una decisión es un vértice, por lo tanto, su tipo es T (el genérico del grafo)
# Decision = T
type Solution[T] = tuple[T, ...]

# 'bab_min_solve' devuelve Optional[ScoredSolution]
type Score = float
type ScoredSolution[T] = tuple[Score, Solution[T]]


# --------------------------------------------------------------------------------

def hamiltoniancycle_bab_solve[T](g: UndirectedGraph[T],
                                  wf: WeightingFunction[T]) -> ScoredSolution[T] | None:
    @dataclass
    class Extra:
        weight: int = 0

    class HamiltonianCycleDS(BabDecisionSequence[T, Extra, Score]):
        def calculate_opt_bound(self) -> Score:
            if self.is_solution():
                return self.extra.weight + wf(self.last_decision(), v_initial)
            if len(self) < len(g.V):
                return self.extra.weight + 0  # Mejorable (arista de menor peso de cada vértice pendiente)
            return infinity

        def calculate_pes_bound(self) -> Score:
            if self.is_solution():
                return self.extra.weight + wf(self.last_decision(), v_initial)
            return infinity  # Mejorable

        def is_solution(self) -> bool:
            return len(self) == len(g.V) and v_initial in g.succs(self.last_decision())

        def successors(self) -> Iterator[Self]:
            if len(self) < len(g.V):
                ds_set = set(self.decisions())  # O(|V|)
                for v in g.succs(self.last_decision()):
                    if v not in ds_set:  # O(1), |V| veces
                        new_weight = self.extra.weight + wf(self.last_decision(), v)
                        yield self.add_decision(v, Extra(new_weight))
                        aa = self.add_decision(v_initial, Extra(0))
                        yield aa

        # Sobreescribimos 'state()'
        def state(self) -> tuple[T, tuple[T, ...]]:
            return self.last_decision(), tuple(sorted(self.decisions()))  # O(|V|)

    initial_ds = HamiltonianCycleDS(Extra(0))
    v_initial = next(iter(g0.V))  # Elegimos un vértice cualquiera como inicial
    one_vertex_ds = initial_ds.add_decision(v_initial, Extra(0))  # Forzamos la primera decisión

    result = bab_min_solve(one_vertex_ds)
    if result is None: return None
    score, solution_ds = result
    return score, solution_ds.decisions()


# Programa principal ---------------------------
if __name__ == "__main__":
    # En este ejemplo, los vértices son de tipo int

    # Por lo tanto, las aristas son tuple[int, int]
    edges = [(0, 2), (0, 3), (0, 9), (1, 3), (1, 4), (1, 8), (2, 3), (2, 5),
             (3, 4), (3, 6), (4, 7), (5, 6), (5, 8), (6, 7), (6, 8), (6, 9)]

    # Y el grafo, g0, es de tipo UndirectedGraph[int]
    g0 = UndirectedGraph(E=edges)

    d0 = dict(((u2, v2), abs(u2 - v2)) for (u2, v2) in g0.E)
    wf0 = WeightingFunction(d0, symmetrical=True)

    sol0 = hamiltoniancycle_bab_solve(g0, wf0)
    if sol0 is not None:
        print(f'Best hamiltonian cycle solution: {sol0}')
    else:
        print('There are no solutions')



from collections.abc import Iterator

from algoritmia.algorithms.traverse import Traverse
from algoritmia.datastructures.graphs import UndirectedGraph

type CC[T] = set[T]  # Connected Component

def connected_components[T](g: UndirectedGraph[T],
                            traverse: Traverse[T]) -> Iterator[CC]:
    pending_vertices = set(g.V)
    while len(pending_vertices) > 0:
        u = pending_vertices.pop()
        cc_vertices = set([v for (u, v) in traverse(g, u)])
        pending_vertices -= cc_vertices
        yield cc_vertices


if __name__ == '__main__':
    from traverse import traverse_bf, traverse_df

    edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)),
             ((2, 2), (2, 3)), ((0, 1), (1, 1)), ((0, 2), (1, 2)),
             ((1, 2), (2, 2)), ((2, 0), (2, 1)), ((0, 3), (1, 3))]

    my_graph = UndirectedGraph(E=edges)

    ccs_bf = list(connected_components(my_graph, traverse_bf))
    print("Breath first:", ccs_bf)

    ccs_df = list(connected_components(my_graph, traverse_df))
    print("Depth first:", ccs_df)



from typing import Self

from algoritmia.schemes.dac_scheme import IDecreaseAndConquerProblem, dec_solve, iter_dec_solve

type Solution = int  # La posición del elemento buscado

class BinarySearchProblem(IDecreaseAndConquerProblem[Solution]):
    def __init__(self, v: list[int], elem: int, start: int, end: int):
        self.v, self.elem, self.start, self.end = v, elem, start, end
        self.mid = (self.start + self.end) // 2

    def is_simple(self) -> bool:
        return self.end - self.start <= 1 or self.elem == self.v[self.mid]

    def trivial_solution(self) -> Solution | None:
        if self.elem == self.v[self.mid]:
            return self.mid
        if self.start == self.end or self.elem != self.v[self.start]:
            return None
        return self.start

    def decrease(self) -> Self:
        if self.elem < self.v[self.mid]:
            return BinarySearchProblem(self.v, self.elem, self.start, self.mid)
        else:
            return BinarySearchProblem(self.v, self.elem, self.mid + 1, self.end)


if __name__ == "__main__":
    my_sorted_list = [2, 3, 3, 11, 12, 18, 21, 29, 30, 37, 43, 75, 82, 98]
    num = 30
    print('list:', my_sorted_list)
    print('number to find:', num)
    problem = BinarySearchProblem(my_sorted_list, num, 0, len(my_sorted_list))

    pos = dec_solve(problem)
    print(f'dec_solve: number found at index {pos}')

    pos = iter_dec_solve(problem)
    print(f'iter_dec_solve: number found at index {pos}')



from typing import Optional

from algoritmia.datastructures.graphs import UndirectedGraph
from algoritmia.datastructures.queues import Fifo

type Vertex = tuple[int, int]


def find_treasure_breadthfirst(g: UndirectedGraph[Vertex],
                               v_start: Vertex,
                               v_treasure: Vertex) -> Optional[Vertex]:
    queue: Fifo[Vertex] = Fifo()
    seen: set[Vertex] = set()
    queue.push(v_start)
    seen.add(v_start)
    while len(queue) > 0:
        v = queue.pop()
        if v == v_treasure:
            return v
        for suc in g.succs(v):
            if suc not in seen:
                seen.add(suc)
                queue.push(suc)
    return None


def find_treasure_depthfirst(g: UndirectedGraph[Vertex],
                             v_start: Vertex,
                             v_treasure: Vertex) -> Optional[Vertex]:
    def explorar_desde(v: Vertex) -> Optional[Vertex]:
        seen.add(v)
        if v == v_treasure:  # preorder
            return v  # preorder
        for suc in g.succs(v):
            if suc not in seen:
                res = explorar_desde(suc)
                if res is not None:
                    return res
        # if v == v_treasure:     # postorder
        #    return v             # postorder

    seen = set()
    return explorar_desde(v_start)


# Main program -------------------------------------------------------------------------

if __name__ == "__main__":
    corridors = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((0, 1), (0, 2)),
                 ((2, 0), (1, 0)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                 ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]
    labyrinth = UndirectedGraph(E=corridors)

    pos_start = (0, 0)
    pos_treasure = (1, 3)

    # room = find_treasure_breadthfirst(labyrinth, pos_start, pos_treasure)
    room = find_treasure_depthfirst(labyrinth, pos_start, pos_treasure)

    if room is None:
        print('Treasure not found.')
    else:
        print(f'Treasure found on room {room}.')



from collections.abc import Iterator
from dataclasses import dataclass
from random import seed, randint
from typing import Self

from algoritmia.schemes.bab_scheme import BabDecisionSequence, bab_max_solve
from algoritmia.schemes.bt_scheme import State

# Tipos  --------------------------------------------------------------------------

Decision = int  # 0 o 1 (coger o no un objeto)
Solution = tuple[Decision, ...]
# Podriamos sobrescribir el método 'solution()' en KnapsackBabDS para que devolviera también el peso.
# En ese caso el tipo de 'Solution' sería:
# Solution = tuple[int, tuple[Decision, ...]]  # (weight, decisions)

# 'bab_max_solve' devuelve Optional[ScoredSolution]
Score = int  # Valor de la mochila (suma del valor de los objetos que contiene)
ScoredSolution = tuple[Score, Solution]


# Versión naif (cotas poco informadas) --------------------------------------------------------------------------

def knapsack_bab_solve_naif(weights: list[int],
                            values: list[int],
                            capacity: int) -> ScoredSolution:
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BabDecisionSequence[Decision, Extra, Score]):
        # OPTIMISTA: Coge TODOS los objetos pendientes (cota poco informada)
        def calculate_opt_bound(self) -> Score:
            return self.extra.value + sum(values[len(self):])

        # PESIMISTA: No coge NINGUNO de los objetos pendientes (cota poco informada)
        def calculate_pes_bound(self) -> Score:
            return self.extra.value + 0

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(values):
                new_weight = self.extra.weight + weights[n]
                if new_weight <= capacity:
                    new_value = self.extra.value + values[n]
                    yield self.add_decision(1, Extra(new_weight, new_value))
                yield self.add_decision(0, self.extra)

        def state(self) -> State:
            return len(self), self.extra.weight

    initial_ds = KnapsackBabDS(Extra())
    score, solution_ds = bab_max_solve(initial_ds)
    return score, solution_ds.decisions()


# Versión con cotas informadas --------------------------------------------------------------------------
c = 0


def knapsack_bab_solve(weights: list[int],
                       values: list[int],
                       capacity: int) -> ScoredSolution:
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BabDecisionSequence[Decision, Extra, Score]):
        # OPTIMISTA: resolver mochila fraccionaria para los objetos que quedan (tema Voraces)
        def calculate_opt_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                f = min(1.0, (capacity - weight) / weights[i])
                weight += f * weights[i]
                value += f * values[i]
                if f < 1:
                    break
            return value

        # PESIMISTA: modificación del optimista para que no fraccione
        def calculate_pes_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                if weight + weights[i] <= capacity:
                    weight += weights[i]
                    value += values[i]
            return value

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterator[Self]:
            n = len(self)
            if n < len(values):
                new_weight = self.extra.weight + weights[n]
                if new_weight <= capacity:
                    new_value = self.extra.value + values[n]
                    yield self.add_decision(1, Extra(new_weight, new_value))
                yield self.add_decision(0, self.extra)

        def state(self) -> State:
            return len(self), self.extra.weight

    initial_ds = KnapsackBabDS(Extra())
    score, solution_ds = bab_max_solve(initial_ds)
    return score, solution_ds.decisions()


# Funciones auxiliares para crear instancias  ----------------------------------------------

def sorted_by_dec_ratio(weights: list[int], values: list[int]) -> tuple[list[int], list[int]]:
    idxs: list[int] = sorted(range(len(weights)), key=lambda i: -values[i] / weights[i])
    w_new = [weights[i] for i in idxs]
    v_new = [values[i] for i in idxs]
    return w_new, v_new


# Crea los problemas con los objetos ordenados de mayor a menor beneficio por kilo
def create_knapsack_problem(num_objects: int) -> tuple[list[int], list[int], int]:
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


# Programa principal -------------------------------------------------------------------------


if __name__ == "__main__":
    # Cada instancia es una tupla (pesos, valores, capacidad):

    # Solution: value = 354, decisions = (1, 0, 1, 0, 0, 0))
    i1 = create_knapsack_problem(6)

    # Solution: value = 1118, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    i2 = create_knapsack_problem(20)

    # Solution: value = 1830, decisions = (1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0...0)
    i3 = create_knapsack_problem(35)

    for W, V, C in [i1, i2, i3]:
        print("-" * 80)
        print(f"Instancia:\n  Pesos = {W}\n  Valores = {V}\n  Capacidad = {C}\n")
        print(f"knapsack_bab_solve:\n  {knapsack_bab_solve(W, V, C)}")
    print("-" * 80)



"""
Version:  5.3 (09-ene-2024)
          5.2 (01-dic-2023)
          5.0 (31-oct-2023)
          4.1 (29-sep-2022)
          4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2024
@license: GPL3
"""
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Iterator, Callable, Sized
from typing import Any, final, Self

# Tipos  --------------------------------------------------------------------------

# El tipo para guardar las secuencias de decisiones (self._decisions) como caminos en el árbol
# de todas las posibles secuencias. Dos secuencias que compartan un prefijo comparten
# la memoria correspondiente a ese prefijo.
type DecisionPath[TDecision] = tuple[TDecision, DecisionPath[TDecision]] | tuple[()]

# ACERCA DEL TIPO State
# La implementación por defecto devuelve self._decisions.
# Podemos sobreescribir state() en la clase hija para devolver otra cosa.
type State = Any

# La clase DecisionSequence -------------------------------------------------------

class DecisionSequence[TDecision, TExtra](ABC, Sized):
    def __init__(self,
                 extra: TExtra | None = None,
                 decisions: DecisionPath[TDecision] = (),
                 length: int = 0):
        self.extra = extra
        self._decisions = decisions
        self._len = length

    # --- Métodos abstractos que hay que implementar en las clases hijas ---

    @abstractmethod
    def is_solution(self) -> bool:
        pass

    @abstractmethod
    def successors(self) -> Iterator[Self]:
        pass

    # --- Método que se puede sobreescribir en las clases hijas: state() ---

    # Debe devolver siempre un objeto inmutable
    # Por defecto se devuelve el contenido de _decisions
    def state(self) -> State:
        return self._decisions

    # -- Métodos finales que NO se pueden sobreescribir en las clases hijas ---

    @final
    def add_decision(self, decision: TDecision, extra: TExtra = None) -> Self:
        return self.__class__(extra, (decision, self._decisions), self._len + 1)

    @final
    def last_decision(self) -> TDecision:  # Es O(1)
        if len(self._decisions) > 0:
            return self._decisions[0]
        raise RuntimeError(f'last_decision() used on an empty {self.__class__.__name__} object')

    @final
    def decisions(self) -> tuple[TDecision, ...]:  # Es O(n)
        ds = deque()
        p = self._decisions
        while p != ():
            ds.appendleft(p[0])
            p = p[1]
        return tuple(ds)

    @final
    def __len__(self) -> int:  # len(objeto) devuelve el número de decisiones del objeto
        return self._len


# Esquema para BT básico --------------------------------------------------------------------------

def bt_solutions[TDecision, TExtra](ds: DecisionSequence[TDecision, TExtra])\
        -> Iterator[DecisionSequence[TDecision, TExtra]]:
    if ds.is_solution():
        yield ds
    for new_ds in ds.successors():
        yield from bt_solutions(new_ds)


#  Esquema para BT con control de visitados --------------------------------------------------------

def bt_vc_solutions[TDecision, TExtra](initial_ds: DecisionSequence[TDecision, TExtra])\
        -> Iterator[DecisionSequence[TDecision, TExtra]]:
    def bt(ds: DecisionSequence[TDecision, TExtra]) -> Iterator[DecisionSequence[TDecision, TExtra]]:
        if ds.is_solution():
            yield ds
        for new_ds in ds.successors():
            new_state = new_ds.state()
            if new_state not in seen:
                seen.add(new_state)
                yield from bt(new_ds)

    seen = {initial_ds.state()}  # marcamos initial_ds como visto
    return bt(initial_ds)  # Devuelve un iterador de soluciones


#  Mejor solución --------------------------------------------------------


def min_solution[TSolution, TScore](solutions: Iterator[TSolution],
                                    f: Callable[[TSolution], TScore]) -> tuple[TScore, TSolution] | None:
    return min(((f(sol), sol) for sol in solutions), default=None)


def max_solution[TSolution, TScore](solutions: Iterator[TSolution],
                                    f: Callable[[TSolution], TScore]) -> tuple[TScore, TSolution] | None:
    return max(((f(sol), sol) for sol in solutions), default=None)



"""

Version: 6.0 (23-oct-2024) - Añade tipo genérico TSolution y cambia Iterator por Iterable
         5.2 (01-dic-2023)
         4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2024
@license: GPL3
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Self


# Divide and conquer -----------------------------------------------------------

class IDivideAndConquerProblem[TSolution](ABC):
    @abstractmethod
    def is_simple(self) -> bool: pass

    @abstractmethod
    def trivial_solution(self) -> TSolution: pass

    @abstractmethod
    def divide(self) -> Iterable[Self]: pass

    @abstractmethod
    def combine(self, solutions: Iterable[TSolution]) -> TSolution: pass


def div_solve[TSolution](problem: IDivideAndConquerProblem[TSolution]) -> TSolution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        subproblems = problem.divide()
        solutions = (div_solve(p) for p in subproblems)
        return problem.combine(solutions)


# Decrease and conquer -----------------------------------------------------------

class IDecreaseAndConquerProblem[TSolution](ABC):
    @abstractmethod
    def is_simple(self) -> bool: pass

    @abstractmethod
    def trivial_solution(self) -> TSolution: pass

    @abstractmethod
    def decrease(self) -> Self: pass

    def process(self, s: TSolution) -> TSolution:
        return s


def dec_solve[TSolution](problem: IDecreaseAndConquerProblem[TSolution]) -> TSolution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        smaller_problem = problem.decrease()
        smaller_solution = dec_solve(smaller_problem)
        return problem.process(smaller_solution)


def tail_dec_solve[TSolution](problem: IDecreaseAndConquerProblem[TSolution]) -> TSolution:
    if problem.is_simple():
        return problem.trivial_solution()
    else:
        return tail_dec_solve(problem.decrease())


def iter_dec_solve[TSolution](problem: IDecreaseAndConquerProblem[TSolution]) -> TSolution:
    while not problem.is_simple():
        problem = problem.decrease()
    return problem.trivial_solution()



from random import seed, randint

type Capacity = int
type Weight = int
type Value = int

type Decision = int  # 0 o 1
type Solution = list[Decision]

type Score = int
type Result = tuple[Score, Solution]  # Siempre habrá solución, no hace falta añadir None

type SParams = tuple[Capacity, int]


# Una solucion son len(v) valores (cada uno puede ser 0 o 1)
# C.E. = O(C N)
# C.T. = O(C N)
def solve(C: Capacity, w: list[Weight], v: list[Value]) -> Result:
    def S(c: Capacity, n: int) -> Score:
        # Caso base
        if n == 0: return 0
        # Recursividad memoizada
        if (c, n) not in mem:
            if w[n - 1] <= c:  # El objeto cabe
                mem[c, n] = max((S(c, n - 1),                       (c, n - 1),    0),
                                (S(c - w[n - 1], n - 1) + v[n - 1], (c - w[n - 1], n - 1), 1))
            else:  # El objeto no cabe
                mem[c, n] = S(c, n - 1), (c, n - 1), 0
        return mem[c, n][0]

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}
    score = S(C, len(v))
    # Recuperamos la solución de mem
    decisions = []
    c, n = C, len(v)
    while n > 0:
        _, (c, n), d = mem[c, n]
        decisions.append(d)
    decisions.reverse()
    return score, decisions


# Una solucion son K valores (los índices de los objetos que metemos en la mochila)
# C.E. = O(C N)
# C.T. = O(C N^2)
def solve2(C: Capacity, w: list[Weight], v: list[Value]) -> Result:
    def S(c: Capacity, n: int) -> Score:
        # Caso base
        if n == 0:
            return 0

        # Recursividad con memoización
        if (c, n) not in mem:
            mem[c, n] = 0, (c, 0), -1  # S no cabe ningún objeto más, quedará c espacio libre
            for d in range(n):
                c_prev, n_prev = c - w[d], d
                if w[d] <= c:
                    mem[c, n] = max(mem[c, n],
                                    (S(c_prev, n_prev) + v[d], (c_prev, n_prev), d))
        return mem[c, n][0]

    mem: dict[SParams, tuple[Score, SParams, Decision]] = {}
    score = S(C, len(v))
    # Recuperación del camino
    c, n = C, len(v)
    decisions = []
    while n > 0:
        _, (c, n), d = mem[c, n]
        if d != -1:
            decisions.append(d)
    decisions.reverse()
    return score, decisions


def sorted_by_dec_ratio(w_old, v_old):
    idxs = sorted(range(len(w_old)), key=lambda i: -v_old[i] / w_old[i])
    w_new = [w_old[i] for i in idxs]
    v_new = [v_old[i] for i in idxs]
    return w_new, v_new


def create_knapsack_problem(num_objects):
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


if __name__ == '__main__':
    w0, v0, C0 = [4, 5, 3, 2, 6], [3, 2, 2, 3, 2], 9  # Solution: value = 8,weight = 9,decisions = (1, 0, 1, 1, 0))
    # w0, v0, C0 = create_knapsack_problem(20)  # Solution: value = 1118, weight = 344, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    # w0, v0, C0 = create_knapsack_problem(35)  # Solution: value = 1830, weight = 543, decisions = (1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    print("Instance:", C0, w0, v0)
    print("Result (traditional X):", solve(C0, w0, v0))
    print("Result (alternative X):", solve2(C0, w0, v0))



import unittest
from algoritmia.algorithms.traverse import traverse_bf, traverse_df
from algoritmia.datastructures.graphs import Digraph
from random import seed


class TestTraversers(unittest.TestCase):
    def setUp(self):
        self.G = Digraph(E=[(0,1), (0,2), (1, 3), (2, 3), (3, 4), (5, 6)])
        seed(0)

    def test_breadth_first_traversal(self):
        self.assertTrue(tuple([v for (u,v) in traverse_bf(self.G, 0)]) in ((0, 1, 2, 3, 4), (0, 2, 1, 3, 4)))

    def test_recursive_preorder_traversal(self):
        self.assertTrue(tuple([v for (u,v) in traverse_df(self.G, 0)]) in ((0, 1, 3, 4, 2), (0, 2, 3, 4, 1)))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


from collections.abc import Iterator
from typing import Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions

# Tipos  --------------------------------------------------------------------------

type Decision = int  # Número de fila donde colocar la reina

# Queremos que una solución sea la secuencia de decisiones (números de fila) en forma de tupla:
type Solution = tuple[Decision, ...]

# - 'bt_solutions' y 'bt_vc_solutions' devuelven un Iterator con las DecisionSequence que
#   llegan a una solución.
# - Pero un objeto DecisionSequence no es una tupla de decisiones: debemos utilizar el método
#   'decisions()' de la clase DecisionSequence para obtener la tupla.

# --------------------------------------------------------------------------------


def nqueens_solutions(board_size: int) -> Iterator[Solution]:
    class NQueensDS(DecisionSequence[Decision, None]):  # Como no hay Extra -> ponemos None
        def is_solution(self) -> bool:
            return len(self) == board_size

        def successors(self) -> Iterator[Self]:
            n = len(self)  # Número de decisiones ya tomadas (reinas colocadas)
            if n < board_size:  # Si quedan decisiones por tomar (reinas por colocar)
                for row in range(board_size):
                    if all(r != row and n - j != abs(row - r)
                           for j, r in enumerate(self.decisions())):
                        yield self.add_decision(row)

    initial_ds = NQueensDS()
    for solution_ds in bt_solutions(initial_ds):
        yield solution_ds.decisions()  # Extraemos las decisiones del objeto solution_ds y las devolvemos


# Programa principal -----------------------------------
if __name__ == "__main__":
    board_size0 = 8

    print(f'n-queens solutions for n={board_size0}:')
    has_solutions = False
    for sol in nqueens_solutions(board_size0):
        has_solutions = True
        print(f'\tSolution: {sol}')
    if not has_solutions:
        print('\tThere are no solutions')



import unittest
from dataclasses import dataclass
from typing import Iterator, Self

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solutions, min_solution


class TestBT_SOLUTIONS(unittest.TestCase):

    def test_4queens_bt_solutions(self):
        class NQueensDS(DecisionSequence[int, None]):
            def is_solution(self) -> bool:
                return len(self) == board_size

            def successors(self) -> Iterator[Self]:
                n = len(self)
                if n < board_size:
                    for row in range(board_size):
                        if all(r != row and n - j != abs(row -r)
                               for j, r in enumerate(self.decisions())):
                            yield self.add_decision(row)

        board_size = 4
        initial_ds = NQueensDS()
        solutions = tuple(s.decisions() for s in bt_solutions(initial_ds))
        self.assertEqual(((1, 3, 0, 2), (2, 0, 3, 1)), solutions)

    def test_coinchange_min_solution(self):
        @dataclass
        class Extra:
            pending: int

        class CoinChangeDS(DecisionSequence[int, Extra]):
            def is_solution(self) -> bool:
                return len(self) == len(v) and self.extra.pending == 0

            def successors(self) -> Iterator[Self]:
                n = len(self)
                if n < len(v):
                    for num_coins in range(self.extra.pending // v[n] + 1):
                        pending2 = self.extra.pending - num_coins * v[n]
                        yield self.add_decision(num_coins, Extra(pending2))

        v, Q = (1, 2, 5, 10), 4
        initial_ds = CoinChangeDS(Extra(Q))
        solutions = (s.decisions() for s in bt_solutions(initial_ds))
        solution = min_solution(solutions, sum)
        self.assertEqual((2, (0, 2, 0, 0)), solution)

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()
